#include "aux.h"
#include "bookmark.h"
#include "version.h"
#include "config.h"
#include "util.h"
extern void error(int code, const char *fmt,...);
extern int save_bookmarks(const char *path, Bookmark *tree){
/* called by quit_cb() in glirc_cb.c */
	static void save_bm(FILE *fp, Bookmark *p);
	FILE *fp;
	time_t t=time(NULL);

	if( (fp=fopen(path, "w"))==NULL) return errno;
	fprintf(fp,sula_NAME sula_VERSION"\nBookmarks created %s\n\n", ctime(&t));
	save_bm(fp, tree);
	fflush(fp);
	fclose(fp);
	return 0;
}
static void save_bm(FILE *fp, Bookmark *p){
	char buf[256];

	if(p){
		save_bm(fp, p->left);
		sprintf(buf,"%s:%s:%s:%u\n", p->channel,
			 p->title?p->title:"", 
			 p->server?p->server:"", 
			 p->port
			 );
		fputs(buf, fp);
	     	save_bm(fp, p->right);
	}
}

#if DEBUG>2
static void    print_bookmark_tree(Bookmark * p)
{

	if (p) {
		print_bookmark_tree(p->right);
		fprintf(stderr,"%d. %s %d %s %s\n",
			 p->id, p->server?p->server:"",
			p->port, p->channel, p->title?p->title:"");
		
		print_bookmark_tree(p->left);
	}
}
#endif

int load_bookmarks(const char *filename, Bookmark **tree ){
	Bookmark *bm;
	FILE *fp;
	char *buf;
	int total=0;
	char *channel, *title, *server, *port;
	#define BUF	512
	if ( !(fp = fopen (filename, "r")) ){
      		return errno;
	} 
       	buf=(char *)malloc(sizeof(char)*(BUF+1));
	if(!buf) return errno;
	while( fgets(buf, BUF, fp)!=NULL){
			char *ptr;
			size_t ll;

			buf[strlen(buf)-1]='\0';
			unspace(buf);
			if(!*buf || (*buf!='#' && *buf!='&')) continue;
			bm=(Bookmark *)malloc(sizeof(Bookmark));
			if(!bm) error(-2,"malloc error");
			ll=strlen(buf);
			channel= (char *)alloca(sizeof(char)*ll);
			
			port=(char *)alloca(sizeof(char)*ll);
			if(!channel || !port) return errno;
			ptr=channel;
			while(*buf && *buf!=':') *ptr++=*buf++;
			if(*buf==':'){ *ptr='\0'; buf++;}
			if(!(*channel)) return -2;
			title= (char *)malloc(sizeof(char)*ll);
			server=(char *)malloc(sizeof(char)*ll);
			ptr=title;
			while(*buf && *buf!=':') *ptr++=*buf++;
			if(*buf==':'){ *ptr='\0'; buf++;}
			if(!(*title)){ free(title); title=NULL;}
			ptr=server;
			while(*buf && *buf!=':') *ptr++=*buf++;
			if(*buf==':'){ *ptr='\0'; buf++;}
			if(!(*server)){ free(server);server=NULL;}
			ptr=port;
			while((*ptr++=*buf++));
			if(*port){
				int i=atoi(port);
				if(i<0) *port='\0';
				ll=strlen(port);
				while ((i=atoi(port))>USHRT_MAX) *(port+ --ll)='\0';
				if( (bm->port=i)==0 ) bm->port= DEFAULT_PORT;
			}
			 else bm->port= DEFAULT_PORT;
			
			bm->channel=channel;
			bm->title=title;
			bm->server=server;
			bm->id=++bm_id;
			insert_bookmark(tree, bm, &total);
			free(title);free(server);
		}
	fclose(fp);
	return 0;
}


static void bm_left_rotate(pBookmark *pp){
	pBookmark	p=*pp, r;
	*pp= r = p->right;
	p->right = r->left;
	r->left=p;
	p->bf--;
	if(r->bf>0) p->bf -= r->bf;
	r->bf--;
	if(p->bf<0) r->bf += p->bf;
}

static void bm_right_rotate(pBookmark *pp){
	pBookmark p=*pp, l;
	*pp= l = p->left;
	p->left = l->right;
	l->right=p;
	p->bf++;
	if(l->bf<0) p->bf -= l->bf;
	l->bf++;
	if(p->bf>0) l->bf += p->bf;
}

extern int insert_bookmark(pBookmark *pp, Bookmark *c, int *id){
/* #if OLD__ then dont allow duplicate bookmarks */

	int dH=0;
	pBookmark p=*pp;
	
	
	assert(c && c->channel);
	if(p==NULL){
		
		*pp=p=(pBookmark)malloc(sizeof(Bookmark));
		if(!p) error(-2,"malloc error");
		p->channel=strdup(c->channel);
		p->title=strdup2(c->title);
		p->server=strdup2(c->server);
		p->port=c->port;
		p->id=c->id;
		p->bf=0;
		p->left=p->right=NULL;
		*id=p->id;
		dH=1;
	}else
{
	/* we shall allow  duplicate bookmarks */
	if( 
	    c->id >p->id
	  ){
	   if(insert_bookmark(&p->right, c, id)){
		p->bf++; 
		if(p->bf==1) dH=1; else
		if(p->bf==2){
			if(p->right->bf==-1) bm_right_rotate(&p->right);
			bm_left_rotate(pp);
		}
	   }
	}else
	if( 
	   c->id<p->id
	  ){
	   if(insert_bookmark(&p->left, c, id)){
	   	p->bf--;
		if(p->bf==-1) dH=1;else
		if(p->bf==-2){
			if(p->left->bf==1) bm_left_rotate(&p->left);
			bm_right_rotate(pp);
		}
	  }
	}
	else
	error(2, "[%s %d] A very bad error occured.\n", __FILE__,__LINE__);
}
	return dH;
}

extern int del_bookmark(pBookmark *pp, Bookmark *c){
	pBookmark p=*pp, *q;
	int dH=0;
	

	
	if(!p) return 0;
	if( 
	    c->id<p->id
	   ){
	   if(del_bookmark(&p->left, c)){
	  	p->bf++;
		if(p->bf == 0) dH=1;else
		if(p->bf == 2){
			if(p->right->bf==-1) bm_right_rotate(&p->right);
			bm_left_rotate(pp);
			if(p->bf==0) dH=1;
		}
	   }
	}else
	if( 
		c->id>p->id
	   ){
	   if(del_bookmark(&p->right, c)){
	   	p->bf--;
		if(p->bf == 0) dH=1;else
		if(p->bf == -2){
			if(p->left->bf==1) bm_left_rotate(&p->left);
			bm_right_rotate(pp);
			if(p->bf==0) dH=1;
		}
	   }
	}else{ /* found bookmark */
		if(p->right ==NULL){
			*pp=p->left;
			free(p);
			 return 1;
		}else
		if(p->left ==NULL){
			*pp=p->right;
			free(p);
			return 1;
		}else{
			q= &p->left;
			while( (*q)->right!=NULL ) q= &(*q)->right;
			/**p=*(*q);*/
			p->channel= strdup((*q)->channel);
			if((*q)->title) p->title=strdup((*q)->title);
			else p->title=NULL;
			p->server=(*q)->server? strdup((*q)->server):NULL;
			p->port=(*q)->port;
			p->id=(*q)->id;
		}

			/**(*q)=*c;*/
			(*q)->channel= strdup(c->channel);
			(*q)->title=c->title?strdup(c->title):NULL;
			(*q)->server=c->server?strdup(c->server):NULL;
			(*q)->port=c->port;
			(*q)->id=c->id;
			if(del_bookmark(&p->left, c)){
				p->bf++;
				if(p->bf==0) dH=1;else
				if(p->bf==2){
					if(p->right->bf==-1) bm_right_rotate(&p->right);
					bm_left_rotate(pp); dH=1;
				}
			}
	}
	return dH;
}

