/* chanwin_cb.c -functions and stuff dealing with channel windows
 * distributed as part of _ s u l a _
 * 
 * Author: Tano Fotang 1997,1998
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License , or (at your option) 
 * any later version.
 * see the file COPYING for more  info.
 * 
 */

#include <forms.h>
#include <sys/time.h>
#include <stdarg.h>
#include "aux.h"
#include "sula.h"
#include "chanwin.h"
#include "etc.h"
#include "global.h"
#include "dcc.h"
#include "ignore.h"
#include "setting.h"
#include "cmd.h"
#include "lag.h"
#include "config.h"
#include "scheme.h"
#include "bookmark.h"

/*static void ClearServerChanList(Winstruct * win);*/

int     subwin_atclose(FL_FORM * form, void *obj)
{
	*((FL_OBJECT **) obj) = NULL;
	return FL_OK;
}
static int     atclose_remove(FL_FORM * form, void *data)
{
	*( (void **)data)=NULL;
	return FL_OK;
}


extern void update_topic(Winstruct * win, const char *channel, const char *topic)
{
	char   *buf;
	
	assert(topic && channel && win);
	if (strcasecmp(channel, win->chanCur->name))
		{return;}
	buf = malloc(sizeof(char) *(strlen(channel)+strlen(topic)+8));
	sprintf(buf, "%s: %s", channel, topic);
	fit_object_label(win->chanwin->WinStatusLine, buf);
	free(buf);
}
void    UpdateStatus(Winstruct * win)
{
	char   *buf;
	int     isserverop;
	if (win->server) {
		int    ischanop = 0,   cantalk = 0;
		char   *chanmode = NULL, *mymode = NULL;
		char   *query = NULL;
		char   *t = NULL, *s = NULL;
		char    has_channels = 0;
		char	*dcc_chat = NULL;
		
		buf = (char *) malloc(sizeof(char) * 1024);
		if (MaxNickLen < 1)
			MaxNickLen = 1;
		if (MaxChanLen < 1)
			MaxChanLen = 1;
		isserverop = (win->server->mode && strchr(win->server->mode, 'o')) ? 1 : 0;
		if (win->server->mode) {
			char *tt=strdchr(win->server->mode, 'o');
			mymode = (char *) malloc(sizeof(char) * (strlen(win->server->mode) + 12));

			if (tt && *tt)
				sprintf(mymode, "  (+%s) ", tt);
			else
				strcpy(mymode, "");
			free(tt);
		}
		else
			mymode = strdup("");
		if (win->chanCur == win->chanStart || win->chanStart->left == win->chanStart) {
			/*  not on a channel (empty channel list) */
			sprintf(buf, "%s%s%s",
				strcp(win->server->nick, MaxNickLen), (isserverop) ? "*" : "",
				mymode);
		}
		else {
			has_channels = 1;
			ischanop = (win->chanCur->modes && strchr(win->chanCur->modes, 'o'));
			cantalk = (win->chanCur->modes && strchr(win->chanCur->modes, 'v'));
			t = (char *) malloc(sizeof(char) * 40);

			if (win->chanCur->limit)
				sprintf(t, " <limit: %d>", win->chanCur->limit);
			else
				strcpy(t, " ");
			if (win->chanCur->key) {
				s = (char *) malloc(sizeof(char) * (strlen(win->chanCur->key) + 10));

				sprintf(s, " <key: %s>", win->chanCur->key);
			}
			if (win->chanCur->modes) {
				char   *tmp = strdup(win->chanCur->modes);
				chanmode = (char *) malloc(sizeof(char) * (
				strlen(win->chanCur->modes) + 10));

				if (ischanop)
					tmp = strdchr(tmp, 'o');
				if (cantalk)
					tmp = strdchr(tmp, 'v');
				if (*tmp)
					sprintf(chanmode, " (+%s) ", tmp);
				else
					strcpy(chanmode, "");
				free(tmp);
			}
			if (win->chanCur->topic)
				update_topic(win, win->chanCur->name, win->chanCur->topic);
			else
				update_topic(win, win->chanCur->name, 
				"No topic is set (" sula_NAME sula_VERSION ")");
		}
		if (fl_get_button(win->chanwin->r_query) && win->query_nick) {
			query = (char *) malloc(sizeof(char) * (strlen(win->query_nick) + 15));

			sprintf(query, " [Query: %s]", win->query_nick);
		}
		else if (	/*
				 * fl_get_button(win->chanwin->r_dcc_chat) && 
				 */
				win->dccCur != win->dccStart &&
				win->dccStart->left != win->dccStart) {
			dcc_chat = (char *) malloc(sizeof(char) * (strlen(win->dccCur->nick) + 15));

			sprintf(dcc_chat, " [DCC chat: %s]", win->dccCur->nick);
		}
		if (has_channels){
			char c[4];
			int i=0;
			if(isserverop) c[i++]=serveropChar;
			if(ischanop) c[i++]=chanopChar;
			if(cantalk && !ischanop) c[i++]=chanvoiceChar;
			c[i]=0;
			sprintf(buf, "%s%s%s%s%s on %s%s%s%s",
				strcp(win->server->nick, MaxNickLen),
				c[0]!=0?c:"", mymode,
				query ? query : "", dcc_chat ? dcc_chat : "",
				strcp(win->chanCur->name, MaxChanLen),
				chanmode ? chanmode : "",
				t ? t : "", s ? s : ""
			   );
		}else{
			char c[2];
			c[0]=0;c[1]=0;
			if(isserverop) t[0]=serveropChar;
			sprintf(buf, "%s%s%s%s%s%s",
				(!query && !dcc_chat) ? "[" sula_NAME sula_VERSION "]  " : "",
				strcp(win->server->nick, MaxNickLen),
				c[0] ? c : "",	mymode, query ? query : "",
				dcc_chat ? dcc_chat : ""
			   );
                }
		free(t);free(mymode); free(dcc_chat); free(query); 
		free(s);free(chanmode);
	}	/*  if( win->server) */
	else if (fl_get_button(win->chanwin->r_dcc_chat) && win->dccCur != win->dccStart &&
		 win->dccStart->left != win->dccStart) {
		buf = (char *) malloc(sizeof(char) * (32+strlen(win->dccCur->nick)));
		sprintf(buf, " [DCC chat with %s]", win->dccCur->nick);

	}
	else{
		srandom(time(NULL));
		buf=strdup(random()%2? "Infoterial: " sula_INFOSITE: "Type /help for help");
	}
	fit_object_label(win->chanwin->status, buf);
	free(buf);
}

void    UpdateWindows(int fd, char connected)
{
/* 
 * update windows status displays upon establishing or closing down a
 * connection */
	WinList *p;
	char    *buf=NULL;
#include "closed.xpm"
#include "connected.xpm"
#include "info.xpm"
#include "here.xpm"
#include "away.xpm"

	if (fd < 0)
		return;
	build_winlistbyfd(fd);
	p = servWin;
	while (p) {
		if (connected) {
			buf = malloc(sizeof(char) * (64+strlen(
			p->win->server->name)+strlen(sula_NAME sula_VERSION)));
			sprintf(buf, "You are now on IRC (via %s, port %d).",
			    p->win->server->name, p->win->server->port);
			fl_set_object_label(p->win->chanwin->WinStatusLine, buf);
			sprintf(buf,"[%d] " sula_NAME sula_VERSION ": %s %d",
			p->win-winstruct+1, p->win->server->name,p->win->server->port);
			fl_set_form_title(p->win->chanwin->chanwin,buf);
			free(buf);
			UpdateStatus(p->win);
			fl_set_pixmapbutton_data(p->win->chanwin->info, info_xpm);
			fl_show_object(p->win->chanwin->info);
			fl_free_pixmapbutton_pixmap(p->win->chanwin->close_con);
			fl_set_pixmapbutton_data(p->win->chanwin->close_con, connected_xpm);
			fl_set_pixmapbutton_data(p->win->chanwin->away,
				 p->win->server->flag & SET_AWAY ? away_xpm : here_xpm);
			fl_show_object(p->win->chanwin->away);

		}
		else {
			CmdSent *cc = cmdsentStart;
			buf = (char *) malloc(sizeof(char) * (64+strlen(
			p->win->server->name?p->win->server->name:p->win->server->servname)));
			sprintf(buf, "Connection to server %s (port %d) closed",
			p->win->server->name?p->win->server->name:p->win->server->servname,
			    p->win->server->port);

			/*fl_addto_browser(p->win->chanwin->brow2, buf);*/
			fl_set_object_label(p->win->chanwin->WinStatusLine, buf);
			free(buf);
			fl_hide_object(p->win->chanwin->info);
			fl_free_pixmapbutton_pixmap(p->win->chanwin->info);

			fl_hide_object(p->win->chanwin->away);
			fl_free_pixmapbutton_pixmap(p->win->chanwin->away);

			fl_free_pixmapbutton_pixmap(p->win->chanwin->close_con);
			fl_set_pixmapbutton_data(p->win->chanwin->close_con, closed_xpm);
			#if 1
			if(p->win->server->connecting>0){
				fl_activate_object(p->win->chanwin->close_con);
				fl_set_object_color(p->win->chanwin->close_con, 
				FL_COL1, FL_MCOL);
			}
			#endif
			if (p->win->chanwin->chanusers->u_vdata) {
				FD_chanusers *f=(FD_chanusers *) p->win->chanwin->chanusers->u_vdata;
				fl_clear_choice(f->channel);
				fl_clear_browser(f->browser);
			}
			while (cc != cmdsentEnd)
				if (cc->win == p->win) {
					CmdSent *q = cc->next;

					free(cc->cmd);
					if (cc->arg)
						free(cc->arg);
					if (q == cmdsentEnd)
						cmdsentEnd = cc;
					else
						*cc = *q;
					free(q);
				}
				else
					cc = cc->next;
		}
		p = p->next;
	}
}


static void free_pups(Winstruct *win){
	fl_freepup(*((int *)win->chanwin->file_menu->u_ldata));
	fl_freepup(*((int *)win->chanwin->file_menu->u_ldata+1));
		 fl_freepup(*((int *)win->chanwin->file_menu->u_ldata+2)); 

		fl_freepup((int)win->chanwin->help_menu->u_ldata);
		fl_freepup(win->chanwin->go_menu->u_ldata);
		fl_freepup(win->chanwin->bookmark_menu->u_ldata);
		fl_freepup(win->chanwin->edit_menu->u_ldata);
		fl_freepup(win->chanwin->window_menu->u_ldata);
}
void del_chanwin_from_list(Winstruct * win)
{
/* destroy a channel window, removing it from the list. if it's the only
 * window on a server, close connection as well. server isnt removed from
 * server list. */
	Go     *p = win->goStart;
	DCCChat *current;
	int     flags = gflags;
	
	gflags &= ~SHOW_TIPS;	/* xforms < 0.87.5: disenable post handler */
	if (win->server) {
		CmdSent *p = cmdsentStart;
		int     x;

		search_winlist_byservname(win->server->servname,
					  win->server->port, &x);
		assert (x);	/* this is fatal error since win!=NULL */
		if (x == 1) {
			if (win->server->fd > -1)
				close_server_connection(&win->server->fd);
			win->server->state = -1;
		}
		else {
			int     err = 0;
			Channel *chan;
			chan = win->chanStart->left;
			while (chan != win->chanStart && err == 0) {
				char    buf[128];

				sprintf(buf, "PART %s\n", chan->name);
				err = sendto_server(win, buf);
				chan = chan->left;
			}
		}
		while (p != cmdsentEnd)
			if (p->win == win) {
				CmdSent *q = p->next;

				free(p->cmd);
				free(p->arg);
				if (q == cmdsentEnd)
					cmdsentEnd = p;
				else
					*p = *q;
				free(q);
			}
			else
				p = p->next;
	}
	
	if (win->chanwin->minput->u_vdata) {
		fl_hide_form( ((FD_minput *)win->chanwin->minput->u_vdata)->minput);
		fl_free_form( ((FD_minput *)win->chanwin->minput->u_vdata)->minput);
	}
	if (win->chanwin->find->u_vdata) {
		fl_hide_form( ((FD_find *)win->chanwin->find->u_vdata)->find);
		fl_free_form( ((FD_find *)win->chanwin->find->u_vdata)->find);
	}
	if (win->chanwin->chanusers->u_vdata) {
		fl_hide_form( ((FD_chanusers *)win->chanwin->chanusers->u_vdata)->chanusers);
		fl_free_form( ((FD_chanusers *)win->chanwin->chanusers->u_vdata)->chanusers);
	}
	current = win->dccStart->left;
	while (current != win->dccStart) {
		remove_dccchat(current->id, NULL);
		current = current->left;
	}
	while (p != win->goEnd) {
		Go     *q = p->next;
		free(p->name);
		if (p->key)
			free(p->key);
		if (q == win->goEnd)
			win->goEnd = p;
		else
			*p = *q;
		free(q);
	}
	free(p);
	ClearWinChanList(win);
	free(win->chanStart);
	free_pups(win);
		
	
	fl_hide_form(win->chanwin->chanwin);
	fl_free_form(win->chanwin->chanwin);
	gflags = flags;
	if (win->query_nick)
		free(win->query_nick);
	winstruct[win-winstruct].chanwin=NULL;
	if(!(gflags&CTRL_WIN_SHOWN)){
	/* show control window if this was the last channel window*/
	   for(flags=0;flags<win_count;flags++)
		if(winstruct[flags].chanwin) break;
	   if(flags==win_count){
		fl_show_form(main_form->main_form, FL_PLACE_FREE, FL_FULLBORDER, 
		sula_NAME sula_VERSION ": Control");
		gflags |= CTRL_WIN_SHOWN;
	   }
	}


}


void    ClearWinChanList(Winstruct * win)
{
	Channel *p, *next;

	assert(win);
	p = win->chanStart->right;
	while (p != win->chanStart) {
		User   *u = p->userStart, *v;
		next=p->right;
		while (u != p->userEnd) {
			free(u->name);
			free(u->address);
			free(u->ircname);
			v = u->next;
			if (v == p->userEnd)
				p->userEnd = u;
			else
				*u = *v;
			free(v);
		}
		free(u);
		p->left->right = p->right;
		p->right->left = p->left;
		free(p);
		p = next;
	}
	win->chanStart->left = win->chanStart->right = win->chanCur = win->chanStart;
}



/* 
 *    CALLBACKS
 */
void    radio_cb(FL_OBJECT * ob, long data)
{
	Winstruct *win = &winstruct[ (int)((FD_chanwin *) ob->form->fdui)->bar->u_ldata];
	switch(data){
	case 0:
		fl_set_button(win->chanwin->r_quote, 1);
		break;
	case 1:
		if (!win->server || win->server->fd<0 || !win->query_nick)
		fl_set_button(win->chanwin->r_echo, 1);
		else fl_set_button(win->chanwin->r_query, 1);
		break;
	case 2: 
		if(win->chanCur == win->chanStart)
		fl_set_button(win->chanwin->r_echo, 1);
		else fl_set_button(win->chanwin->r_channel, 1);
		break;
	case 3: 
		if(win->dccCur == win->dccStart)
		fl_set_button(win->chanwin->r_echo, 1);
		else fl_set_button(win->chanwin->r_dcc_chat, 1);
		break;
	case 4:
		break;
	}
	UpdateStatus(win);
}

void    join_cb(FL_OBJECT * ob, long data)
{
/* join channels */
	Winstruct *win = &winstruct[ (int)((FD_chanwin *) ob->form->fdui)->bar->u_ldata];
	char *buf=NULL;
	if (!win->server){
		fl_addto_browser(win->chanwin->brow2, PROMPT "No server connection");
		return;
	}
	switch (fl_mouse_button()) {
	    case FL_LEFTMOUSE:
		    jjoin(win);
		    return;

	    case FL_MIDDLEMOUSE:
		    if (win->lastPart->name) {
			    buf = (char *)malloc(sizeof(char)*(strlen(
			    win->lastPart->name)+ (win->lastPart->key ? 
			    strlen(win->lastPart->key):0)+10));
			    sprintf(buf, "JOIN %s %s", win->lastPart->name,
			      win->lastPart->key ? win->lastPart->key : "");
		    }
		    break;
	    default:
		    if (win->server->lastInvite) {
			    buf = (char *)malloc(sizeof(char)*(strlen(
			    win->server->lastInvite)+8));
			    sprintf(buf, "JOIN %s", win->server->lastInvite);
			    
		    }
		    break;
	}
	if(buf){
	       process_cmd(buf, win);
	       free(buf);
	}
}

void    jjoin(Winstruct * win)
{
	char   *buf = NULL;
	char   *argv[10];
	int i=-1, j=0;
	static char already_open = 0;

	if (already_open || !win->server)
		return;
	else {
		char    *s=(char *) malloc(sizeof(char) *(56+strlen(win->server->name)));

		already_open = 1;
		sprintf(s, "Type channel names below, seperated by blanks."
			);
		buf = getstr2(s, sula_NAME sula_VERSION ": Join channels", "", 0);
		free(s);
	}
	buf = clean(buf);
	if (!buf)
		goto out;
	buf = strtok(buf, ", ");
#define MAX_JOIN 10
	while (++i < MAX_JOIN && buf) {
		argv[i] = (char *)malloc(sizeof(char)*(strlen(buf) + 8));
		if (!argv[i])
			goto out;
		if (*buf != '#' && *buf != '&')
			sprintf(argv[i], "join #%s", buf);
		else
			sprintf(argv[i], "join %s", buf);
		buf = strtok(NULL, ", ");
		/*process_cmd( argv[i], win)      doesnt work: buf becomes NULL! */
	}
	while (j < i && !process_cmd(argv[j], win)) free(argv[j++]);
      out:
	free(buf);
	already_open = 0;

}
void    PartChannel_cb(FL_OBJECT * ob, long data)
{
	Winstruct *win = &winstruct[ (int)((FD_chanwin *) ob->form->fdui)->bar->u_ldata];
	if (!win->server)
		return;
	if (win->server->fd < 0) {
		if (gflags & BEEP_ERR) fl_ringbell(0);
		return;
	}
	switch (fl_mouse_button()) {
	    case FL_LEFTMOUSE:
		    if (win->chanCur == win->chanStart) {
			    if (gflags & BEEP_ERR) fl_ringbell(0);
			    return;
		    }
		    else {
			    char *buf = (char *)malloc(sizeof(char)*(strlen(
			    win->chanCur->name)+8));

			    sprintf(buf, "part %s", win->chanCur->name);
			    process_cmd(buf, win);
			    free(buf);
		    }
		    break;
	    case FL_MIDDLEMOUSE:{
			    Channel *current;
			    char    buf[128];

			    current = win->chanStart->right;
			    while (current != win->chanStart) {
				    sprintf(buf, "part %s", current->name);
				    process_cmd(buf, win);
				    current = current->right;
				    delay(1000);
			    }
		    }
		    break;
	    default:
		    if (win->lastJoin->name) {
			    char *buf = (char *)malloc(sizeof(char)*(strlen(
			    win->lastJoin->name)+8));
			    sprintf(buf, "part %s", win->lastJoin->name);
			    process_cmd(buf, win);
			    free(buf);
		    }
		    else {
			    if (gflags & BEEP_ERR) fl_ringbell(0);
			    return;
		    }
		    break;
	}
}
void    dcc_cb(FL_OBJECT * ob, long data)
{
	
	Winstruct *win = &winstruct[ (int)((FD_chanwin *) ob->form->fdui)->bar->u_ldata];
	long m=fl_mouse_button();
	switch (m) {
	    case FL_LEFTMOUSE:	/*accept last chat request in current window */
	    case FL_MIDDLEMOUSE: /*accept last chat request using a new chat window */
		    if (chatreqStart == chatreqEnd)
			    break;
		    else {
			    char    buf[16];

			    sprintf(buf, "achat %d", chatreqStart->id);
			    process_cmd(buf, m==FL_MIDDLEMOUSE?NULL:win);
		    }
		    break;

	    default:		/*close current char */
		    if (win->dccCur == win->dccStart)
			    return;
		    else {
			    char *buf = (char *)malloc(sizeof(char)*(strlen(win->dccCur->nick)+
			    (win->dccCur->email ? strlen(win->dccCur->email) : 1)+64));
			    sprintf(buf, "Closing DCC Chat connection to %s (%s)",
				    win->dccCur->nick, win->dccCur->email ? win->dccCur->email : "?");
			    fit_object_label(win->chanwin->WinStatusLine, buf);
			    free(buf);
			    if (!remove_dccchat(win->dccCur->id, NULL)) {
				    fl_addto_browser(win->chanwin->brow2, 
				    PROMPT "Error closing dcc chat");
				    return;
			    }
			    UpdateStatus(win);
		    }
		    return;
	}

}

static void minput_send(FL_OBJECT * ob)
{
	FD_minput *form = (FD_minput *) ob->form->fdui;
	Winstruct *win = (Winstruct *) ob->u_vdata;
	char   *buf;

	buf = (char *) fl_get_input(form->mline);
	if (!(*buf))
		return;
	
	
	if (fl_get_button(win->chanwin->r_quote)) {
		#if defined _GUILE
		parse_scheme(buf,win-winstruct);
		#else
		int     errur = 0;
		char *ptr;

		if (!(gflags & QUOTE_ML)) {	/* dont automatically quote to server:flooding! */
			if (gflags & BEEP_ERR) fl_ringbell(0);
			fl_addto_browser(win->chanwin->brow2, 
			PROMPT "quote to server? set quote_ml to on");
			return;
		}
		if (!win->server)
			return;
		ptr = strtok(buf, "\n");
		while (ptr && !errur) {
			char    *buff=(char *)malloc(sizeof(char)*(strlen(ptr)+3));

			sprintf(buff, "%s\n", ptr);
			errur = sendto_server(win, buff);
			free(buff);
			ptr = strtok(NULL, "\n");
		}

		#endif
		return;
	}
	else {
		char   *buf2=NULL;
		int     notice = fl_get_button(form->r_notice);

		if (fl_get_button(win->chanwin->r_dcc_chat)) {
			if (win->dccCur == win->dccStart) {
				if (gflags & BEEP_ERR) fl_ringbell(0);
				fl_addto_browser(win->chanwin->brow2, PROMPT "No DCC conection!");
				return;
			}
			fmt_string = "%n%c%h%p$";
			buf2 = format_msg(fmt_you_dccchat,
					  win->server ? win->server->nick : Name, win->dccCur->nick,
					  win->dccCur->email ? win->dccCur->email : win->dccCur->nick, buf);
			if (buf2)
				fl_addto_browser(win->chanwin->browser, buf2);
		}
		else if (fl_get_button(win->chanwin->r_query))
			if (win->query_nick) {
				fmt_string = "%n%c%p$";
				buf2 = format_msg(notice ? fmt_you_priv_notice : fmt_you_priv_msg,
				   win->server->nick, win->query_nick, buf);
				if (buf2)
					fl_addto_browser(win->chanwin->browser, buf2);
			}
			else {
				if (gflags & BEEP_ERR) fl_ringbell(0);
				fl_addto_browser(win->chanwin->brow2, PROMPT "No nick to query!");
				return;
			}
		else if (fl_get_button(win->chanwin->r_channel)) {
			if (win->chanCur == win->chanStart) {
				if (gflags & BEEP_ERR) fl_ringbell(0);
				fl_addto_browser(win->chanwin->brow2, PROMPT "No on any channel");
				return;
			}
			fmt_string = "%n%c%p$";
			buf2 = format_msg(notice ? fmt_you_pub_notice : fmt_you_pub_msg,
				win->server->nick, win->chanCur->name,buf);
			if (buf2)
				fl_addto_browser(win->chanwin->browser, buf2);

		}else{
		/* echo */
		 	fl_addto_browser(win->chanwin->browser, buf);
			return;
		}
		free(buf2);
		buf2=NULL;
		/* send msg now */
				
		if (fl_get_button(win->chanwin->r_query))
			if (win->query_nick) {
				char *tt=
				ctcp_quote_it(buf, strlen(buf));
				buf2 = (char *) malloc(sizeof(char) * (strlen(tt) +
				strlen( win->query_nick) + 16));
				sprintf(buf2, "%s %s :%s\n", notice ? "NOTICE" : "PRIVMSG", win->query_nick, tt);
				sendto_server(win, buf2);
				free(tt);
			}
			else {
				fl_addto_browser(win->chanwin->brow2, PROMPT "No nick for query!");
				return;
			}
		else if (fl_get_button(win->chanwin->r_channel)) {
			char *tt=
			ctcp_quote_it(buf, strlen(buf));
			buf2 = (char *) malloc(sizeof(char) * (strlen(tt) +
				strlen( win->chanCur->name) + 16));
			sprintf(buf2, "%s %s :%s\n", notice ? "NOTICE" : "PRIVMSG", win->chanCur->name, tt);
			sendto_server(win, buf2);
			free(tt);
		}
		else {
			if (fl_get_button(win->chanwin->r_dcc_chat)) {
				char *tt=ctcp_quote_it(buf, strlen(buf));
				buf2 = (char *) malloc(sizeof(char) * (strlen(tt) +2));
				sprintf(buf2, "%s\n", tt);
				sendto_dcc_chat(win, win->dccCur, buf2);
				free(tt);
			}
			else {
				fl_addto_browser(win->chanwin->brow2, PROMPT "(bug?) no destination");
				return;
			}
		}
		free(buf2);
	}
}
void    minput_cb(FL_OBJECT * ob, long data)
{

	if (ob->u_vdata) {
		fl_raise_form(((FD_minput *)ob->u_vdata)->minput);
		return;
	} else{
	Winstruct *win = &winstruct[ (int)((FD_chanwin *) ob->form->fdui)->bar->u_ldata];
	FD_minput *form= create_form_minput();
	ob->u_vdata=(void *)form;
	fl_set_form_atclose(form->minput, atclose_remove, (void *) &ob->u_vdata);
	fl_set_object_prehandler(form->mline, set_cursor);
	form->mline->u_ldata = (long) XC_xterm;
	form->send->u_vdata = (void *) win;
	form->close->u_vdata = (void *) &ob->u_vdata;
	form->clear->u_vdata = (void *) form->mline;
	fl_set_form_minsize(form->minput, 385, 185);
	fl_show_form(form->minput, FL_PLACE_MOUSE | FL_FREE_SIZE,
		FL_TRANSIENT, sula_NAME sula_VERSION ": Multiline message");
	fl_do_forms();
	}

}

void    minput_cb_cb(FL_OBJECT * ob, long data)
{
	switch (data) {
	    case 0:		/*send */
		    minput_send(ob);
		    break;
	    case 1:		/* clear */
		    fl_set_input((FL_OBJECT *) ob->u_vdata, "");
		    break;

	    case 2:{		/*dismiss */
			    FD_minput *form = (FD_minput *) ob->form->fdui;

			    fl_hide_form(form->minput);
			    fl_free_form(form->minput);
			    *( (void **)ob->u_vdata) = NULL;
			    break;
		    }
	    default:
		    break;
	}

}

void    menu_cb(FL_OBJECT * ob, long data)
{
	static void do_file_menu(Winstruct *);
	static void do_edit_menu(Winstruct *);
	static void do_bookmark_menu(Winstruct *);
	Winstruct *win;
	
	win = &winstruct[ (int)((FD_chanwin *) ob->form->fdui)->bar->u_ldata];
	switch (data) {
	    case 0:
		    do_file_menu(win);
		    break;
	    case 1:{		/* go */
		int     item = fl_get_menu(ob);

		    if (item <= 0)
			    break;
		    else {
			    char   *chan = (char *) fl_get_menu_item_text(ob, item);
			    Go     *p = win->goStart;

			    while (p != win->goEnd)
				    if (!strcasecmp(p->name, chan)) {
					    char    *buf;
					    
					    buf=(char *)malloc(sizeof(char)*(strlen(p->name)+64));
					    sprintf(buf, "join %s %s", p->name, p->key ? p->key : "");
					    process_cmd(buf, win);
					    free(buf);
					    break;
				    }
				    else
					    p = p->next;
		    }
		    break;
		  }
	    case 2:
		    do_bookmark_menu(win);
		    break;
	    case 3:
		    {
			    static void do_ignore_win(Winstruct *);
			    static void do_dcc_win(Winstruct *);
			    static void do_bookmark_win(Winstruct *);
			    int     item = fl_get_menu(ob);

			    if (item <= 0)
				    break;
			    switch(item){
			    case 1:
				    do_dcc_win(win);break;
			    case 2:
				    do_ignore_win(win);break;
			    case 3:
				    fl_trigger_object(win->chanwin->chanusers);
				    break;
			    case 4:
				    do_bookmark_win(win);break;
			    case 5:
			            fl_trigger_object(main_form->servers);
				    break;
			    default:
				    if (gflags & BEEP_ERR) fl_ringbell(0);
				    break;
                            }
			    break;
		    }
	    case 4:
		    {
			    int     item = fl_get_menu(ob);

			    if (item <= 0)
				    return;
			    if (item == 1)
				    do_about();
			    else if (item == 2)
				    do_help();
			    else
				    if (gflags & BEEP_ERR) fl_ringbell(0);
		    }
		    break;
	    case 5:		/*edit */
		    {
			    do_edit_menu(win);
			    break;
		    }
	    default:
		    break;
	}			/* switch */
}
static void find_text(FL_OBJECT * ob)
{
	FD_find *form = (FD_find *) ob->form->fdui;
	Winstruct *win = (Winstruct *) ob->u_vdata;
	int     total, *i = (int *) ob->u_ldata;
	char   *text;
	int     backwards = 0;
	static int previous = 0;

	text = (char *) fl_get_input(form->input);
	if (!(*text)) {
		if (gflags & VERBOSE_CLIENT)
			fl_addto_browser(win->chanwin->brow2, PROMPT "find: search for what??");
		if (gflags & BEEP_ERR) fl_ringbell(0);
		return;
	}
#if 0
	*((char **) form->done->u_vdata) = (char *)text;
#endif
	total = fl_get_browser_maxline(win->chanwin->browser);
	if (total < 1) {
		*i = 1;
		if (gflags & VERBOSE_CLIENT)
			fl_addto_browser(win->chanwin->brow2, PROMPT "find: empty browser");
		if (gflags & BEEP_ERR) fl_ringbell(0);
		return;
	}
	backwards = fl_get_button(form->direction);
	if (!backwards && *i < 1)
		*i = 1;
	if (!fl_get_button(form->case_sense))
		text = lowercase(text);

	for (; (!backwards && *i <= total) || (backwards && *i >= 1);) {
		char   *s;

		s = (char *) fl_get_browser_line(win->chanwin->browser, *i);
		if (!s || !(*s))
			goto checkpos;
		if (!fl_get_button(form->case_sense))
			s = lowercase(s);
		if (strstr(s, text)) {
			if (previous)
				fl_deselect_browser_line(win->chanwin->browser, previous);
			fl_select_browser_line(win->chanwin->browser, *i);
			previous = *i;
			fl_set_browser_topline(win->chanwin->browser, *i);
			if (!backwards)
				(*i)++;
			else
				(*i)--;
			return;
		}
	      checkpos:
		if (backwards)
			(*i)--;
		else
			(*i)++;
	}
	if (*i > total)
		(*i)--;
	if (gflags & VERBOSE_CLIENT) {
		if (!backwards)
			fl_addto_browser(win->chanwin->brow2,
					 PROMPT "find: Reached end of document. No matching string found");
		else
			fl_addto_browser(win->chanwin->brow2,
					 PROMPT "find: Reached start of document. No matching string found");
		if (gflags & BEEP_ERR) fl_ringbell(0);
	}
}

static void find(Winstruct * win)
{
	static char *s =0;
	static int i = 1;
	FD_find *form;

	if (win->chanwin->find->u_vdata) {
		fl_raise_form(((FD_find *)win->chanwin->find->u_vdata)->find);
		return;
	}
	form=create_form_find();
	win->chanwin->find->u_vdata=(void *)form;
	fl_set_form_atclose(form->find, atclose_remove, (void *) &win->chanwin->find->u_vdata);
	fl_set_object_prehandler(form->input, set_cursor);
	form->input->u_ldata = (long) XC_xterm;
	fl_set_input(form->input, s);
	form->go->u_vdata = (void *) win;
	form->go->u_ldata = (long) &i;	/* current brow line */
	form->done->u_vdata = (void *) &s;
	form->direction->u_vdata=(void *)&win->chanwin->find->u_vdata; /* done */
	form->clear->u_vdata = (void *) form->input;
	fl_set_form_minsize(form->find, 400, 160);
	fl_set_form_maxsize(form->find, 600, 160);
	fl_show_form(form->find, FL_PLACE_MOUSE | FL_FREE_SIZE,
	      FL_TRANSIENT, sula_NAME sula_VERSION ": Find line with text");
	fl_do_forms();

}
void    find_cb(FL_OBJECT * ob, long data)
{
	switch (data) {
	    case 0:		/*send */
		    find_text(ob);
		    break;
	    case 1:		/* clear */
		    fl_set_input((FL_OBJECT *) ob->u_vdata, "");
		    break;

	    case 2:{		/*dismiss */
		FD_find *form = (FD_find *) ob->form->fdui;
		/*Winstruct *win=(Winstruct *)form->go->u_vdata;*/
		*((int *)form->go->u_ldata)=1;
		fl_hide_form(form->find);
		fl_free_form(form->find);
		*((void **)form->direction->u_vdata) = NULL;
		break;
		    }
	    default:
		    break;
	}

}

static int lost_clipboard_cb(FL_OBJECT * ob, long type)
{
	return 0;
}

static int get_clipboard_cb(FL_OBJECT * ob, long type, const void *data, long size)
{
	if (size == 0) {
		if (gflags & BEEP_ERR) fl_ringbell(0);
		if (gflags & VERBOSE_CLIENT)
			fl_addto_browser(ob, PROMPT "clipboard is empty");
	}
	else
		fl_addto_browser(ob, (char *) data);
	return 0;
}


static int get_send_clipboard_cb(FL_OBJECT * ob, long type, const void *data, long size)
{
	if (size == 0) {
		if (gflags & BEEP_ERR) fl_ringbell(0);
		return -1;
	}
	else 
		send_blocktxt(
		             ((FD_chanwin *) ob->form->fdui)->bar->u_ldata,
		             (char *)data,
			     1);
	return 0;
}

void send_blocktxt(int win, const char *text, int flag)
{
/* send text containin newlines. almost like multiline msg except hthat msg
 * is sent line after line here 
 * flag: add to cmd log*/
	char   *ptr;
	char   *buf = (char *) text;

	if (*buf==0)
		return;
	while (buf) {
		ptr = strsep(&buf, "\n");
		parse_input(win, ptr, flag);
	}
}
static void edit_browser(FL_OBJECT * browser, int cut)
{
	FD_chanwin *form = (FD_chanwin *) browser->form->fdui;
	int     total = fl_get_browser_maxline(browser);
	int     i;
	char   *buf = 0;
	size_t  ll = 0;

	fl_freeze_form(form->chanwin);
	for (i = 1; i <= total; i++)
		if (fl_isselected_browser_line(browser, i)) {
			char   *s = (char *) fl_get_browser_line(browser, i);
			buf = realloc(buf, (ll + strlen(s) + 2) * sizeof(char));
			if (!buf) {
				error(-1,"malloc error");
				return;
			}
			sprintf(buf + ll, "%s\n", s);
			ll = strlen(buf);
			if (cut) {
				fl_delete_browser_line(browser, i);
				total--;
				i--;
			}
		}
	fl_unfreeze_form(form->chanwin);
	if (buf) {
		*(buf + --ll) = '\0';	/* remove last \n */
		fl_stuff_clipboard(browser, FL_STRING, (void *) buf, ll, lost_clipboard_cb);
	}
	else {
		if (gflags & BEEP_ERR) fl_ringbell(0);
		if (gflags & VERBOSE_CLIENT)
			fl_addto_browser(form->brow2,
					 PROMPT "No block selcted");
	}
}

static void delete_lines(FL_OBJECT * browser)
{
	FD_chanwin *chanwin = (FD_chanwin *) browser->form->fdui;
	int     total = fl_get_browser_maxline(browser);
	int     i;

	fl_freeze_form(chanwin->chanwin);
	for (i = 1; i <= total; i++)
		if (fl_isselected_browser_line(browser, i)) {
			fl_delete_browser_line(browser, i);
			total--;
			i--;
		}
	fl_unfreeze_form(chanwin->chanwin);
}

static char *extract_nick(const char *str){
	size_t ll=strcspn(str, "!");
	char *s=malloc(sizeof(char)*(1+ll));
	strncpy(s,str,ll);
	*(s+ll)='\0';
	return s;
}
static char *sender2nick(const Sender *s){
	
	return (extract_nick(s->sender));
}
static void do_edit_menu(Winstruct * win)
{
	int     item = fl_get_menu(win->chanwin->edit_menu);

	if (item <= 0)
		return;
	switch (item) {
	    case 11:
		    edit_browser(win->chanwin->browser, 0);
		    break;	/* copy browser 1 */
	    case 21:
		    edit_browser(win->chanwin->brow2, 0);
		    break;
	    case 12:
		    edit_browser(win->chanwin->browser, 1);
		    break;	/* cut */
	    case 22:
		    edit_browser(win->chanwin->brow2, 1);
		    break;
	    case 13:
		    fl_request_clipboard(win->chanwin->browser,
					 FL_STRING, get_clipboard_cb);
		    break;
	    case 23:		/* paste into browser */
		    fl_request_clipboard(win->chanwin->brow2,
					 FL_STRING, get_clipboard_cb);
		    break;

	    case 14:
		    delete_lines(win->chanwin->browser);
		    break;
	    case 24:
		    delete_lines(win->chanwin->brow2);
		    break;	/*delete */
	    case 15:
	    case 25:		/* unselect all */
		    fl_freeze_form(win->chanwin->chanwin);
		    fl_deselect_browser(item < 20 ? win->chanwin->browser : win->chanwin->brow2);
		    fl_unfreeze_form(win->chanwin->chanwin);
		    break;
	    case 3:		/* paste into input line */
		    fl_request_clipboard(win->chanwin->clock, FL_STRING, get_send_clipboard_cb);
		    break;
	    default:
		    break;
	}
}

	
void    chanwin_cb(FL_OBJECT * ob, long data)
{
	int id=((FD_chanwin *) ob->form->fdui)->bar->u_ldata;
	Winstruct *win = &winstruct[id];
	switch (data) {

	    case 0:		/* sinput */
		    parse_input(id, fl_get_input(ob),
		    	1/*fl_get_button(win->chanwin->history)*/);
		    if(win->chanwin) /* some cammands might have destroyed window */
		    	fl_set_input(ob, "");
		    break;

	    case 41:		/* cut */
		    switch (fl_mouse_button()) {
			case FL_MIDDLEMOUSE:
				edit_browser(win->chanwin->brow2, 1);
				break;
				break;
			default:
				edit_browser(win->chanwin->browser, 1);
				break;
				break;
		    }
		    break;

	    case 40:		/* copy */
		    switch (fl_mouse_button()) {
			case FL_MIDDLEMOUSE:
				edit_browser(win->chanwin->brow2, 0);
				break;
				break;
			default:
				edit_browser(win->chanwin->browser, 0);
				break;
				break;
		    }
		    break;

	    case 42:		/* delete */
		    switch (fl_mouse_button()) {
			case FL_MIDDLEMOUSE:
				delete_lines(win->chanwin->brow2);
				break;
				break;
			default:
				delete_lines(win->chanwin->browser);
				break;
				break;
		    }
		    break;
	    case 43:		/* paste */
		    switch (fl_mouse_button()) {
			case FL_LEFTMOUSE:
				fl_request_clipboard(win->chanwin->clock, FL_STRING, get_send_clipboard_cb);
				break;
			case FL_MIDDLEMOUSE:
				fl_request_clipboard(win->chanwin->browser, FL_STRING, get_clipboard_cb);
				break;
			default:
				fl_request_clipboard(win->chanwin->brow2, FL_STRING, get_clipboard_cb);
				break;
		    }
		    break;
	    case 4:{		/* find */
			    find(win);
			    break;
		    }
	    case 5:		/* info */
		    break;

	    case 6:{		/* away */
			    char   *msg = (char *)fl_get_input(win->chanwin->sinput);
			    char   *buf;

			    if (!(win->server->flag&SET_AWAY)) {
				    if (!(*msg))
					    msg = away_text;
			    }
			    else if (!(*msg))
				    msg = NULL;
			    if (!msg)
				    buf = strdup("away");
			    else {
				    buf = (char *)malloc(sizeof(char)* (strlen(msg) + 8));
				    sprintf(buf, "away %s", msg);
				    fl_set_input(win->chanwin->sinput, "");
			    }
			    if (process_cmd(buf, win))
				    fl_addto_browser(win->chanwin->brow2, PROMPT "away failed!");
			    free(buf);
			    break;
		    }
	    case 7:		/* (dis)connect */
		    if (!win->server) {
			    fl_addto_browser(win->chanwin->brow2,
					     PROMPT "Window has no server. Use /server to connect");
			    if (gflags & BEEP_ERR) fl_ringbell(0);
		    }
		    else if (win->server->fd < 0)
			    connect_to_server(win->server, win,NULL,0);
		    else
			    new_process_cmd("quit", id);
		    break;

	    case 8:		/* topic */
		    if (win->chanCur == win->chanStart) {
			    if (gflags & BEEP_ERR) fl_ringbell(0);
			    return;
		    }
		    else {
			    char   *cmd = NULL;
			    long    m = fl_mouse_button();

			    if (m == FL_MIDDLEMOUSE)
				    cmd = strdup("topic -clear");
			    else if (m == FL_RIGHTMOUSE)
				    cmd = strdup("topic");
			    else {
				    char   *str = (char *) fl_get_input(win->chanwin->sinput);

				    if (str)
					    cmd = (char *) malloc(sizeof(char) * (
					    strlen(str) + strlen(win->chanCur->name) + 10));
				    else
					    cmd = (char *) malloc(sizeof(char) * (strlen(win->chanCur->name) + 10));
				    sprintf(cmd, "topic %s %s", win->chanCur->name, str ? str : "");
			    }
			    if (cmd){
				    process_cmd(cmd, win);
				    free(cmd);
				}
			    fl_set_input(win->chanwin->sinput, "");
		    }
		    break;
	    case 9:		/*nick*/
	    	if(!win->server || win->server->fd<0){
			if (gflags & BEEP_ERR) fl_ringbell(0);
			if(gflags&VERBOSE_CLIENT)  fl_addto_browser(win->chanwin->brow2,
					     PROMPT"No server connection");
			break;
		}
		switch(fl_mouse_button()){
			case FL_LEFTMOUSE:{/* nick */
				char *s=nextstr((char *)fl_get_input(win->chanwin->sinput),1);
				if(!s || *s==0){
				      goto erred;	
				}else{
				char *buf= (char *) malloc(sizeof(char) * (strlen(s)+8));
				sprintf(buf,"nick %s",s);
				process_cmd(buf, win);
				fl_set_input(win->chanwin->sinput, "");
				free(buf); free(s);
				}
			}
			return;
			case FL_MIDDLEMOUSE:{	/* user mode*/
				char *s=(char *)fl_get_input(win->chanwin->sinput);
				char *buf;
				if(s==NULL || !(*s)) goto erred;
				buf = (char *) malloc(sizeof(char) * (strlen(s)+8)+
					strlen(win->server->nick));
				sprintf(buf,"mode %s %s",win->server->nick, s);
				process_cmd(buf, win);
				fl_set_input(win->chanwin->sinput, "");
				free(buf);
				return;
			}
			default:{	/* chan mode*/
				if(win->chanStart==win->chanCur){
				if (gflags & BEEP_ERR) fl_ringbell(0);
				if(gflags&VERBOSE_CLIENT)  fl_addto_browser(win->chanwin->brow2,
					     PROMPT"change mode: no channel");
				}else{
				 char *s=(char *)fl_get_input(win->chanwin->sinput);
				 char *buf;
				 if(!(*s)) goto erred;
				 buf = (char *) malloc(sizeof(char) * (strlen(s)+8)+strlen(win->chanCur->name));
				sprintf(buf,"mode %s %s",win->chanCur->name, s);
				process_cmd(buf, win);
				fl_set_input(win->chanwin->sinput, "");
				free(buf);
				}
				return;
			}
		}
		erred:
			if (gflags & BEEP_ERR) fl_ringbell(0);
			if(gflags&VERBOSE_CLIENT) 
				fl_addto_browser(win->chanwin->brow2,
				    PROMPT"arguments needed on command line");
			return;
	    case 11:		/* lag */
		    #ifndef NOLAG
		    if(!win->server || win->server->fd<0){
		    	fl_set_object_label(win->chanwin->WinStatusLine,
				    "You are not connected to a server");
		   }else  send_lag_junk(win->server,1,1);
		   #endif
		    break;
	    case 10:		/* go! */
	    case 12:{		/* do it! */
			    char   *target;
			    char   *cmd, *buf;
			    //char history=1;

			    target = (char *) fl_get_choice_text(win->chanwin->target);
			    if (target==0 || *target==0) {
				    if (gflags & BEEP_ERR) fl_ringbell(0);
				    fl_addto_browser(win->chanwin->brow2, PROMPT "No target!");
				    return;
			    }
			    cmd = (char *) fl_get_choice_text(win->chanwin->operation);
			    if (cmd==NULL || !(*cmd)) {
				    if (gflags &BEEP_ERR)  fl_ringbell(0);
				    fl_addto_browser(win->chanwin->brow2, PROMPT "No operation?");
				    return;
			    }
			    target=extract_nick(target);
			    if (data == 12) {	/* we dont need a msg */
				    buf = (char *) alloca(sizeof(char) * (strlen(cmd) + strlen(target) + 5));
				    sprintf(buf, "%c%s %s", cmdchar, cmd, target);
			    }
			    else {
				    char   *msg = (char *) fl_get_input(win->chanwin->sinput);

				    if (*msg==0)
					    {
					   if (gflags &BEEP_ERR)  fl_ringbell(0);
				    fl_addto_browser(win->chanwin->brow2, PROMPT "empty input line");
					    return;
					    }
				    buf = alloca(sizeof(char)* (strlen(msg) + strlen(cmd) + 
				    strlen(target) + 5));
			            sprintf(buf, "%c%s %s %s", cmdchar, cmd, target, msg);
				    fl_set_input(win->chanwin->sinput, "");
			    }
			    if ((gflags & HISTORY) && !AddToHistory(buf))
					win->chanwin->cmd_fwd->u_ldata = (long) 0L;

			    if (new_process_cmd(buf+1, id)) {
				    buf = (char *)alloca(sizeof(char)*(strlen(cmd) + 32));
				    sprintf(buf, PROMPT "%s: cant find command", cmd);
				    fl_addto_browser(win->chanwin->brow2, buf);
			    }
			    free(target);
			    break;
		    }		/* case 10 */
	    case 13:		/* update targets */
		    {
			Sender *s;
			if(senderStart==senderStart->left)
				{fl_ringbell(0);break;}
			fl_clear_choice(win->chanwin->target);
			s=senderStart->left;
			while(s!=senderStart){
				fl_addto_choice(win->chanwin->target,s->sender);
				s=s->left;
			}
		    }
		    break;
	    
	    case 14:
	    if(fl_get_button(win->chanwin->history))
	         sflags |= LOG;
	    else
	         sflags &= ~LOG;
	    break;
	    default:
		    {fprintf(stderr,"[%s %d] forgot %ld\n",__FILE__,__LINE__,data);break;}

	    case 30:{		/* cmd fwd */
			    History *hist = ((History *) ob->u_ldata);

			    if (hist == NULL)
				    hist = histCur;
			    else
				    hist = hist->right;
			    ob->u_ldata = (long) hist;
			    if (hist == histStart && (hist = hist->right) == histStart)
				    return;	/*empty list */
			    fl_set_input(win->chanwin->sinput, hist->line);
		    }
		    break;
	    case 31:{
			    History *hist = (History *) ((FD_chanwin *) ob->form->fdui)->cmd_fwd->u_ldata;

			    if (hist == NULL)
				    hist = histCur;
			    else
				    hist = hist->left;
			    ((FD_chanwin *) ob->form->fdui)->cmd_fwd->u_ldata = (long) hist;
			    if (hist == histStart && (hist = hist->left) == histStart)
				    return;	/*empty list */
			    fl_set_input(win->chanwin->sinput, hist->line);
		    }
		    break;
	    case 32:	/* tab key ;  */
	    {
	     		Sender *s=((Sender *) ob->u_ldata);
			if (s == NULL)
				s = senderCur;
			else
			    s = s->right;
			ob->u_ldata = (long) s;
			if (s == senderStart && (s = senderStart->right) == senderStart)
				{fl_ringbell(0);break;}	/*empty list */
			else{
				char *buf, *nick=sender2nick(s);
				char *cmd = (char *) fl_get_choice_text(win->chanwin->operation);
			    	assert(cmd && *cmd);
				buf=malloc(sizeof(char)*(strlen(cmd)+strlen(nick)+5));
				sprintf(buf, "%c%s %s ", cmdchar, cmd, nick);
				fl_set_input(win->chanwin->sinput, buf);
				free(nick); free(buf);
			}


	    }
		    break;
	    case 33:{		/* chan_right */
			    if (fl_get_button(win->chanwin->r_channel)) {
				    win->chanCur = win->chanCur->right;
				    if (fl_mouse_button() == FL_MIDDLEMOUSE)
					    win->dccCur = win->dccCur->right;
				    if (win->dccCur == win->dccStart)
					    win->dccCur = win->dccCur->right;
				    if (win->chanCur == win->chanStart &&
					(win->chanCur = win->chanCur->right) == win->chanStart
					&& win->dccCur == win->dccStart)
					    return;	/* empty channel list */
			    }
			    else if (fl_get_button(win->chanwin->r_dcc_chat)) {
				    win->dccCur = win->dccCur->right;
				    if (fl_mouse_button() == FL_MIDDLEMOUSE)
					    win->chanCur = win->chanCur->right;
				    if (win->chanCur == win->chanStart)
					    win->chanCur = win->chanCur->right;
				    if (win->dccCur == win->dccStart &&
					(win->dccCur = win->dccCur->right) == win->dccStart &&
					win->chanCur == win->chanStart
				       )
					    return;	/* no dcc chat connections, no channels */
			    }
			    else {
				    /* if(fl_get_button(win->chanwin->r_channel)) */
				    win->chanCur = win->chanCur->right;
				    if (win->chanCur == win->chanStart &&
					(win->chanCur = win->chanCur->right) == win->chanStart)
					    return;	/* empty channel list */
			    }
			    UpdateStatus(win);
			    break;
		    }
	    case 34:{		/* chan_left */

			    if (fl_get_button(win->chanwin->r_channel)) {
				    win->chanCur = win->chanCur->left;
				    if (fl_mouse_button() == FL_MIDDLEMOUSE)
					    win->dccCur = win->dccCur->left;
				    if (win->dccCur == win->dccStart)
					    win->dccCur = win->dccCur->left;
				    if (win->chanCur == win->chanStart &&
					(win->chanCur = win->chanCur->left) == win->chanStart &&
					win->dccCur == win->dccStart
				       )
					    return;	/* empty channel list */
			    }
			    else if (fl_get_button(win->chanwin->r_dcc_chat)) {
				    win->dccCur = win->dccCur->left;
				    if (fl_mouse_button() == FL_MIDDLEMOUSE)
					    win->chanCur = win->chanCur->left;
				    if (win->chanCur == win->chanStart)
					    win->chanCur = win->chanCur->left;
				    if (win->dccCur == win->dccStart &&
					(win->dccCur = win->dccCur->left) == win->dccStart &&
					win->chanCur == win->chanStart
				       )
					    return;	/* no dcc chat connections */
			    }
			    else {
				    /* if(fl_get_button(win->chanwin->r_channel)) */
				    win->chanCur = win->chanCur->left;
				    if (win->chanCur == win->chanStart &&
					(win->chanCur = win->chanCur->left) == win->chanStart)
					    return;	/* empty channel list */
			    }
			    UpdateStatus(win);
			    break;
		    }

	}			/* switch */
}

void    chanusers_cb(FL_OBJECT * ob, long data)
{
	
	
	Winstruct *win = &winstruct[ (int)((FD_chanwin *) ob->form->fdui)->bar->u_ldata];
	switch(fl_mouse_button()){
	
	case FL_RIGHTMOUSE: 
		if (win->chanCur != win->chanStart) {
			char    *buf=malloc(sizeof(char)*(strlen(win->chanCur->name)
			+16));

			sprintf(buf, "who -update %s", win->chanCur->name);
			new_process_cmd(buf, win-winstruct);
			free(buf);
		}
		return;
	case FL_MIDDLEMOUSE:
		new_process_cmd("lu", win-winstruct);
		return;
   default:{

	Channel *chan;
	FD_chanusers *form;
	if (ob->u_vdata) {
		fl_raise_form( ((FD_chanusers *)ob->u_vdata)->chanusers);
		return;
	}
	form = create_form_chanusers();
	ob->u_vdata=(void *)form;
	fl_set_form_atclose(form->chanusers,atclose_remove, (void *) &ob->u_vdata);

	form->input->u_ldata = (long) XC_xterm;
	fl_set_object_prehandler(form->input, set_cursor);
	fl_set_input_return(form->input, FL_RETURN_END);

	fl_set_object_bw(form->base, -2);
	fl_set_object_bw(form->bar, -2);
	fl_set_object_bw(form->button_grp, -2);
	form->quit->u_vdata = (void *) &ob->u_vdata;
	form->update->u_vdata = (void *) win;
	form->channel->u_vdata = (void *) win;
	form->others->u_vdata = (void *) win;
	form->op->u_vdata = form->deop->u_vdata = (void *) win;
	form->msg->u_vdata = form->chat->u_vdata = (void *) win;
	form->do_it->u_vdata = (void *) win;
	form->query->u_vdata = (void *) win;

	fl_set_button(form->kick, 0);
	fl_set_button(form->ban, 0);
	fl_set_button(form->deop, 0);

	fl_addto_choice(form->others,
		"clientinfo|finger|ping|time|version|userinfo|chelp|whois");
	fl_set_choice(form->others, 2);
	fl_set_choice_fontsize(form->others, 12);
	/*fl_set_choice_align(form->others,FL_ALIGN_LEFT); */
	fl_set_choice_fontsize(form->channel, FL_MEDIUM_SIZE);
	fl_set_choice_align(form->channel, FL_ALIGN_LEFT);
	fl_set_choice_fontstyle(form->channel, FL_TIMESITALIC_STYLE);

	chan = win->chanStart->left;
	while (chan != win->chanStart) {
		fl_addto_choice(form->channel, chan->name);
		chan = chan->left;
	}
	fl_show_form(form->chanusers, FL_PLACE_MOUSE, FL_TRANSIENT,
		     sula_NAME sula_VERSION ": Channel users");
	if (win->chanCur != win->chanStart)
		load_userlist(win, win->chanCur->name, 0);
	fl_do_forms();
	return;
	}
  }/* switch*/

}

void    chanusers_cb_cb(FL_OBJECT * ob, long data)
{
	FD_chanusers *form = (FD_chanusers *) ob->form->fdui;

	switch (data) {
	    case 4:{		/* others */
		char   *cmd = (char *) fl_get_choice_text(form->others);
		int     total, i, no_sel = 1;

		if (!cmd || !(*cmd))
		    return;
		total = fl_get_browser_maxline(form->browser);
		    for (i = 1; i <= total; i++)
			    if (fl_isselected_browser_line(form->browser, i)) {
				    char   *nick,*line = nextstr((char *)fl_get_browser_line(form->browser, i), 1);
				    char   *arg=NULL;
				    no_sel = 0;
				    nick = nextstr(line, 1);
				    arg = (char *) malloc(sizeof(char) * (strlen(cmd) + strlen(nick) + 2));
				    sprintf(arg, "%s %s", cmd, nick+2); /* skip over @f */
				    free(nick);free(line);
				    if (process_cmd(arg, (Winstruct *) ob->u_vdata))
					    break;
				    fit_object_label(form->status, arg);
				    free(arg);
			    }
			    if (no_sel)
				    fit_object_label(form->status, "no selection?");
		    }
		    break;
	    case 7:{		/* cahnnel */
		    const char *channel = fl_get_choice_text(form->channel);

		    if (!channel)
			    return;
		    else
			    load_userlist((Winstruct *) ob->u_vdata, channel, 0);
		    }
		    break;
	    case 1:
	    case 0:{
		    const char *channel = fl_get_choice_text(form->channel);
		    int     total, i;
		    char    mode[] = "-o", no_sel = 1, *nick, *buf;
		    if (!channel || !(*channel))
			    return;
		    if (data == 0)
			    strcpy(mode, "+o");
			    total = fl_get_browser_maxline(form->browser);
			    for (i = 1; i <= total; i++)
				    if (fl_isselected_browser_line(form->browser, i)) {
					    nick = nextstr((char *)fl_get_browser_line(form->browser, i), 1);
					    buf=malloc(sizeof(char)*(strlen(nick)+strlen(channel)+16));
					    no_sel = 0;
					    sprintf(buf, "MODE %s %s %s\n", channel, mode, nick + 2);
					    free(nick);
					    if (sendto_server((Winstruct *) ob->u_vdata, buf))
						    break;
						free(buf);
				    }
			    if (no_sel)
				    fit_object_label(form->status, "selection?");
			    else
				    fit_object_label(form->status, "op/deop sent...");
		    }
		    break;
	    case 2:{		/*query */
		    int     total, i;
		   char    no_sel = 1;
		    total = fl_get_browser_maxline(form->browser);
		    for (i = 1; i <= total; i++)
			    if (fl_isselected_browser_line(form->browser, i)) {
				    char   *nick = nextstr((char *)fl_get_browser_line(form->browser, i), 1);
				    char   *arg;

				    no_sel = 0;
				    arg = (char *) malloc(sizeof(char) * (strlen(nick) + 8));

					    sprintf(arg, "query %s", nick+2);
					    free(nick);
					    if (process_cmd(arg, (Winstruct *) ob->u_vdata))
						    fit_object_label(form->status, "errur..");
					    else
						    fit_object_label(form->status, arg);
					   free(arg);
					    break;
				    }
			    if (no_sel)
				    fit_object_label(form->status, "no selection?");
		    }
		    break;
	    case 3:{		/* chat */
	    	    char   *cmd = (char *) fl_get_choice_text(form->others);
	    	    int     total, i;
	    	    char    no_sel = 1;

	    	    if (!(*cmd))
	    		    return;
	    	    total = fl_get_browser_maxline(form->browser);
	    	    for (i = 1; i <= total; i++)
	    		    if (fl_isselected_browser_line(form->browser, i)) {
	    			    char   *nick = nextstr((char *)fl_get_browser_line(form->browser, i), 1);
	    			    char   *arg;

	    			    no_sel = 0;
	    			    arg = (char *) malloc(sizeof(char) * (strlen(nick) + 8));
	    			    sprintf(arg, "chat %s", nick+2);
				    free(nick);
	    			    if (process_cmd(arg, (Winstruct *) ob->u_vdata))
	    				    break;
	    				free(arg);
	     		    }
	     	    if (no_sel)
	     		    fit_object_label(form->status, "dest?");
	     	    else
	     		    fit_object_label(form->status, "ok");
		    }
		    break;
	    case 5:{		/* update */
	     	    char   *chan = (char *) fl_get_choice_text(form->channel);
	     	    if (!chan)
	     		    return;
	     	    else {
	     		    char *buf=(char *)malloc(sizeof(char)*(strlen(chan)+16));

	     		    sprintf(buf, "WHO -update %s", chan);
	     		    process_cmd(buf, (Winstruct *) ob->u_vdata);
			    free(buf);
	     		    fit_object_label(form->status, "wait for update..");
	     	    }
		    }
		    break;
	    case 6:{		/* dismiss */
	     	    fl_hide_form(form->chanusers);
	     	    fl_free_form(form->chanusers);
		    *((void **)ob->u_vdata)=NULL;
	     	    break;
		    }
	    case 8:{		/* do it */
	     	    Winstruct *win = (Winstruct *) ob->u_vdata;
	     	    char    kick, ban, ig;
	     	    char *msg = NULL;
	     	    const char *channel = NULL;
	     	    int     total = fl_get_browser_maxline(form->browser);
	     	    int     i, no_sel = 1;

	     	    kick = fl_get_button(form->kick);
	     	    ban = fl_get_button(form->ban);
	     	    ig = fl_get_button(form->ignore);
	     	    if (!kick && !ban && !ig) {
	     		    fit_object_label(form->status, "failed. do what?");
	     		    return;
	     	    }
		    if (kick)
	     		    msg = (char *)fl_get_input(form->input);
	     	    if (ban || kick) {
	     		    channel = fl_get_choice_text(form->channel);
	     		    if (!channel) {
	     			    fit_object_label(form->status, "failed. channel?");
	     			    return;
	     		    }
	     	    }
	     	    for (i = 1; i <= total; i++)
	     		    if (fl_isselected_browser_line(form->browser, i)) {
	     			    char   *nick = nextstr((char *)fl_get_browser_line(form->browser, i), 1);
	     			    char   *addr = nextstr((char *)fl_get_browser_line(form->browser, i), 3);
				    char   *buf=NULL;
	     			    no_sel = 0;
	     			    if (ban) {
					    buf=(char *)malloc(sizeof(char)*(strlen(channel)+
					    strlen(addr)  +32));
					    if(*addr=='~') strshift(addr,1);
	     				    sprintf(buf, "mode %s +b *!*%s", channel,addr); 
					    process_cmd(buf, win);
				    	    free(buf);
	     			    }
	     			    if (kick) {
					    buf=(char *)malloc(sizeof(char)*(strlen(channel)+
					    strlen(nick)+(msg ?strlen(msg):strlen(
					    sula_NAME sula_VERSION)) +16));
					    if(!buf) return;
					    sprintf(buf, "kick %s %s %s", channel, nick + 2, 
					    msg ? msg : sula_NAME sula_VERSION);
					    process_cmd(buf, win);
				    		free(buf);
	     			    }
	     			    if (ig) {
					    buf=(char *)malloc(sizeof(char)*(strlen(addr)+16));
	     				    sprintf(buf, "ignore *!*%s", addr);
					    process_cmd(buf, win);
				    	    free(buf);
	     			    }
				    free(nick);free(addr);
	     		    }
	     	    if (no_sel)
	     		    fit_object_label(form->status, "destination?");
	     	    else
	     		    fit_object_label(form->status, "done..");

		    }
		    break;
	    case 9:{		/*msg */
	    	    char *msg = (char *)fl_get_input(form->input);
	    	    int     total, i;
	    	    char    dest[1024], *buf;
	    	    size_t  ll;
	    	    Winstruct *win;
	    	    if (!(*msg))
	    		    return;
	    	    win = (Winstruct *) ob->u_vdata;
		    ll = strlen(msg);
	    	    dest[0] = '\0';
	    	    total = fl_get_browser_maxline(form->browser);
	    	    for (i = 1; i <= total; i++)
	    		    if (fl_isselected_browser_line(form->browser, i)) {
	    			    char   *nick = nextstr((char *)fl_get_browser_line(form->browser, i), 1);

	    			    strshift(nick,2);
	    			    buf = (char *)malloc(sizeof(char)*(ll + strlen(nick) + 32));
	    			    strcat(dest, nick);
	    			    strcat(dest, " ");
	    			    sprintf(buf, "PRIVMSG %s :%s\n", nick, msg);
	    			    free(nick);
				    if (sendto_server(win, buf))
	    				    break;
					free(buf);
	    		    }

	    	    if (dest[0]) {
	    		    fmt_string = "%n%c%p$";
	    		    buf = format_msg(fmt_you_priv_msg, win->server->nick, dest, msg);
	    		    if (buf)
	    			    fl_addto_browser(win->chanwin->browser, buf);
	    		    fl_set_input(form->input, "");
	    		    fit_object_label(form->status, "message done");
	    	    }
	    	    else
	    		    fit_object_label(form->status, "selected?");

		    }
	    default:
		    break;
	}			/* switch */
}				/* end of function */

#define LOAD	0
#define SAVE	1
#define save_  "Save as:",sula_NAME sula_VERSION ": Save browser lines"
#define load_ "File name:",sula_NAME sula_VERSION ": Load a file"
static void save_load_browser(Winstruct * win, char what, char save_all)
{
		
	char   *ptr;

	if(what==SAVE) ptr = getstr2(save_, NULL, 1);
	else ptr = getstr2(load_, NULL, 1);
	if (!ptr)
		return;
	else {
		char *buf=malloc(sizeof(char)*(strlen(ptr)+32));
		sprintf(buf,"window -%s 1 -f %s%s",what==LOAD?"l":"s",
		 ptr, (what==SAVE && save_all)?"":" -m");
		process_cmd(buf, win);
		free(buf);
	}
}

static void do_file_menu(Winstruct * w)
{
	static void newwin(Winstruct * w);
	static void load_hist(Winstruct *, int);
	
	/*static void load_file(Winstruct *);*/
	int     item;

	if ((item = fl_get_menu(w->chanwin->file_menu)) <= 0)
		return;
	switch (item) {
	    case 1:
		    newwin(w);
		    break;
	    case 2:
		    fl_clear_browser(w->chanwin->browser);
		    break;
	    case 3:
		    fl_clear_browser(w->chanwin->brow2);
		    break;
	    case 401:
		    save_load_browser(w,SAVE, 1);
		    break;
	    case 402:
		    save_load_browser(w,SAVE, 0);
		    break;
	    case 5:
		    save_load_browser(w, LOAD, 0);
		    break;
	    case 701:		/* clear */
		    process_cmd("history -clear", w);
		    break;
	    case 702:		/*reload */
		    process_cmd("history -reload", w);
		    break;
	    case 703:
		    load_hist(w, 1);
		    break;
	    case 704:
		    load_hist(w, 0);
		    break;
		    process_cmd("history -save", w);
		    break;
	    case 8:
		    if(w->chanwin->close_con->u_vdata){
		    /* have requested server connection. Wait for it! */
		    	fl_addto_browser(w->chanwin->brow2,PROMPT"Cannot close."
			"Waiting for server connection. Waiting.");
			fl_ringbell(0);
		    }else
		    if(
		    w->chanStart==w->chanCur ||
		    confirm("Leave channels and close window?",
	 		sula_NAME sula_VERSION": Confirm", 7)
			)
		    del_chanwin_from_list(w);
		    break;
	    case 9:
		    fl_trigger_object(main_form->quit);
		    break;
	    default:
		    break;
	}
}
static void load_hist(Winstruct * win, int load)
{
	char   *ptr, *buf;

	if (!(ptr = getenv("HOME")))
		ptr = ".";
	buf = (char *)malloc(sizeof(char)* (strlen(ptr) + 8 + strlen(P_HOME P_HISTORY)));
	sprintf(buf, "%s/%s/%s", ptr, P_HOME, P_HISTORY);
	ptr = getstr2("Type the name of the history file below",
		      sula_NAME sula_VERSION ": Load history file", buf, 1);
	ptr = clean(ptr);
	if (!ptr)
		free(buf);
	else {
		free(buf);
		buf = (char *)malloc(sizeof(char)* (strlen(ptr) + 24));
		if (load)
			sprintf(buf, "history -load %s", ptr);
		else
			sprintf(buf, "history -save %s", ptr);
		process_cmd(buf, win);
		free(buf);
	}
}




static void newwin(Winstruct * w)
{
/* create a channel window */
	Winstruct *win;
	char   *label=NULL;

#include "info.xpm"
#include "here.xpm"
#include "away.xpm"
#include "connected.xpm"

	win=new_win();
	if(!win) return;
	if (w->server) {
		win->server = (Server *) malloc(sizeof(Server));
		if (win->server == NULL)
                   error(-2, "malloc error");
		win->server=w->server;
		label = (char *) malloc(sizeof(char) *
			(strlen(win->server->name?
			win->server->name:win->server->servname)+
			strlen(sula_NAME sula_VERSION) + 32));
		if(label)
		sprintf(label, "[%d] "sula_NAME sula_VERSION ": %s %d",
		(win-winstruct)+1, win->server->name?
			win->server->name:win->server->servname, win->server->port);
		if (win->server->fd > -1)
		  if(!win->server->connecting || 
		  	win->chanwin->close_con->u_vdata /*ext. process looking up hostname*/)
			{
			fl_set_pixmapbutton_data(win->chanwin->info, info_xpm);
			fl_show_object(win->chanwin->info);
			fl_free_pixmapbutton_pixmap(win->chanwin->close_con);
			fl_set_pixmapbutton_data(win->chanwin->close_con, connected_xpm);
			fl_set_pixmapbutton_data(win->chanwin->away, 
			win->server->flag&SET_AWAY ? away_xpm : here_xpm);
			fl_show_object(win->chanwin->away);

		}else{
			fl_deactivate_object(win->chanwin->close_con);
			fl_set_object_color(win->chanwin->close_con, 
				FL_INACTIVE_COL, FL_MCOL);
		}
	}
	else {
	      win->server=NULL;
		label = (char *) malloc(sizeof(char) * (strlen(sula_NAME sula_VERSION) + 32));
		if(label)
		sprintf(label, "[%d] "sula_NAME sula_VERSION 
			": No server connection", (win-winstruct)+1);
	}
	show_chanwin(win, &label,1,1);

}
void show_chanwin(Winstruct *win, char **s, int show_form, int do_form){
	if(show_form){
		if(*s==NULL){
		*s=malloc(sizeof(char) * (strlen(sula_NAME sula_VERSION) + 32));
		sprintf(*s, "[%d] "sula_NAME sula_VERSION 
			": No server connection", (win-winstruct)+1);
		}
		fl_show_form(win->chanwin->chanwin,
		FL_PLACE_FREE | FL_FREE_SIZE,
		FL_FULLBORDER, *s);
	}
	UpdateStatus(win);
	if(do_form){
		fl_do_forms();
	}
}
static void file_menu_setup(FL_OBJECT * menu);
static void bookmarks_setup(FL_OBJECT * menu);
static void window_menu_setup(FL_OBJECT * menu);
static void help_menu_setup(FL_OBJECT * menu);
static void edit_menu_setup(FL_OBJECT * menu);

void    menu_setup(Winstruct * win)
{
/* called by global.c */
	edit_menu_setup(win->chanwin->edit_menu);
	window_menu_setup(win->chanwin->window_menu);
	help_menu_setup(win->chanwin->help_menu);
	bookmarks_setup(win->chanwin->bookmark_menu);
	file_menu_setup(win->chanwin->file_menu);
}


static void help_menu_setup(FL_OBJECT * menu)
{
	int p;

	if(menu->u_ldata!=-1) fl_freepup(menu->u_ldata);
retry:
	p = fl_newpup(FL_ObjWin(menu));
	if(p==-1){
		inc_pup_max();
		goto retry;
	}
	fl_setpup_shadow(p, 0);
	fl_setpup_cursor(p, XC_right_ptr);
	fl_setpup_bw(p, -1);
	fl_addtopup(p, "About sula%x1%h%l", "#A#a");
	fl_addtopup(p, "User Guide%x2%h", "#U#u");
	fl_set_menu_popup(menu, p);
	menu->u_ldata=(long)p;/* to free it when destroying channel window */
}

static void window_menu_setup(FL_OBJECT * menu)
{
	int p;
	if(menu->u_ldata!=-1) fl_freepup(menu->u_ldata);
retry:
	p = fl_newpup(FL_ObjWin(menu));
	if(p==-1){
		inc_pup_max();
		goto retry;
	}
	fl_setpup_shadow(p, 0);
	fl_setpup_cursor(p, XC_right_ptr);
	fl_setpup_bw(p, -1);
	fl_addtopup(p, "DCC window\tAlt-D%x1%h%i", "#D#d");
	fl_addtopup(p, "Ignore\t\tAlt-I%x2%h", "#I#i");
	fl_addtopup(p, "Channel users\tAlt-C%x3%h", "#C#c");
	fl_addtopup(p, "Bookmarks\tAlt-B%x4%h", "#B#b");
	fl_addtopup(p, "Servers\t\tAlt-S%x5%h", "#S#s");
	fl_set_menu_popup(menu, p);
	menu->u_ldata=(long)p;/* to free pupup when destroying channel window */
}
static void edit_menu_setup(FL_OBJECT * menu)
{
	int p;
	if(menu->u_ldata!=-1) fl_freepup(menu->u_ldata);
retry:
	p = fl_newpup(FL_ObjWin(menu));
	if(p==-1){
		inc_pup_max();
		goto retry;
	}
	fl_setpup_shadow(p, 0);
	fl_setpup_cursor(p, XC_right_ptr);
	fl_setpup_bw(p, -1);
	fl_addtopup(p, "Browser 1%i");
	fl_addtopup(p, "Copy%x11%h", "Cc");
	fl_addtopup(p, "Cut%x12%h", "uU");
	fl_addtopup(p, "Paste%x13%h", "Pp");
	fl_addtopup(p, "Delete%x14%h", "Dd");
	fl_addtopup(p, "Unselect all%x15%l");
	fl_addtopup(p, "Browser 2%i");
	fl_addtopup(p, "Copy	Alt-C%x21%h", "#C#c");
	fl_addtopup(p, "Cut	Alt-U%x22%h", "#u#U");
	fl_addtopup(p, "Paste	Alt-P%x23%h", "#P#p");
	fl_addtopup(p, "Delete	Alt-D%x24%h", "#D#d");
	fl_addtopup(p, "Unselect all%x25%l");
	fl_addtopup(p, "Paste%x3%h", "aA");	/* paste into input line */
	fl_set_menu_popup(menu, p);
	menu->u_ldata=p;
}

/*----------------windows_menu setup--------------------------------------*/

static int ignore_form_atclose(FL_FORM * form, void *obj)
{
	*((FD_ignore **) obj) = NULL;
	return FL_OK;
}

static void do_ignore_win(Winstruct * win)
{
	char    *buf;

	if (ignore_form) {
		fl_raise_form(ignore_form->ignore);
		return;
	}
	ignore_form = create_form_ignore();
	fl_set_form_atclose(ignore_form->ignore, ignore_form_atclose, (void *) &ignore_form);

	/*fl_deactivate_object(ignore_form->load); */
	buf= (char *) malloc(sizeof(char) *(66+32));
	sprintf(buf, "%-66s time left", "ignored mask");
	fit_object_label(ignore_form->title, buf);
	free(buf);
	ignore_form->done->u_vdata = (void *) &ignore_form;
	fl_show_form(ignore_form->ignore, FL_PLACE_MOUSE, FL_TRANSIENT, sula_NAME sula_VERSION ": Ignore list");
	fl_call_object_callback(ignore_form->update);
	fl_do_forms();

}

static void do_dcc_win(Winstruct * win)
{

#if 0
	static char already_open = 0;
	static int do_files = 1;

	if (already_open) {
		if (gflags & BEEP_ERR)  fl_ringbell(0);
		return;
	}
	else
		already_open = 1;
	dcc_form = create_form_dcc();
	fl_set_form_atclose(dcc_form->dcc, atclose_reset, (void *) &already_open);
	dcc_form->chat_brow->u_ldata = dcc_form->file_brow->u_ldata = (long) XC_right_ptr;
	fl_set_object_prehandler(dcc_form->file_brow, set_cursor);
	fl_set_object_prehandler(dcc_form->chat_brow, set_cursor);
	if (do_files)
		fl_set_button(dcc_form->r_file, 1);
	else
		fl_set_button(dcc_form->r_chat, 1);
	dcc_form->done->u_vdata = (void *) &already_open;
	dcc_form->r_file->u_ldata = (long) &do_files;
	dcc_form->r_chat->u_ldata = (long) &do_files;
	fl_show_form(dcc_form->dcc, FL_PLACE_MOUSE, FL_FULLBORDER, sula_NAME sula_VERSION ": DCC connections");
#define XXX_
#ifndef XXX_
	fl_hide_object(dcc_form->chat_brow);
#else

	if (do_files) {
		fl_hide_object(dcc_form->chat_brow);

		fl_show_object(dcc_form->file_brow);
		fl_set_button(dcc_form->r_file, 1);
	}
	else {
		fl_hide_object(dcc_form->file_brow);
		fl_show_object(dcc_form->chat_brow);
		fl_set_button(dcc_form->r_chat, 1);
	}
#endif
	fl_do_forms();
#endif

}
static void file_menu_setup(FL_OBJECT * menu)
{
	int     *p;
	p=malloc(sizeof(int)*3);
	if(menu->u_ldata!=-1) fl_freepup(menu->u_ldata);
 
retry2:
	p[0] = fl_newpup(FL_ObjWin(menu));
	if(p[0]==-1){
		inc_pup_max();
		goto retry2;
	}
retry3:
	p[1] = fl_newpup(FL_ObjWin(menu));
	if(p[1]==-1){
		inc_pup_max();
		goto retry3;
	}
retry:
	p[2] = fl_newpup(FL_ObjWin(menu));
	if(p[2]==-1){
		inc_pup_max();
		goto retry;
	}
	fl_setpup_shadow(p[2], 0);
	fl_setpup_cursor(p[2], XC_right_ptr);
	fl_setpup_bw(p[2], -1);
	fl_setpup_shadow(p[0], 0);
	fl_setpup_bw(p[0], -1);
	fl_setpup_shadow(p[1], 0);
	fl_setpup_bw(p[1], -1);

	fl_addtopup(p[2], "New window\tAlt-N%x1%h%l", "#N#n");

	fl_addtopup(p[2], "Clear browser1\tAlt-C%x2%h", "#C#c");
	fl_addtopup(p[2], "Clear browser2\tAlt-B%x3%h%l", "#b#B");
	fl_addtopup(p[2], "Save browser1\tAlt-S%x4%h%m", "#S#s", p[0]);
	fl_addtopup(p[0], "Everything\tAlt-E%x401%h", "#E#e");
	fl_addtopup(p[0], "Selected lines\tAlt-S%x402%h", "#S#s");
	fl_addtopup(p[2], "Load file\t\tAlt-L%x5%h%l", "#L#l");

	fl_addtopup(p[2], "History\t\tAlt-H%x7%h%m%l", "#H#h", p[1]);
	fl_addtopup(p[1], "Clear\t\tAlt-C%x701%h", "#C#c");
	fl_addtopup(p[1], "Reload\t\tAlt-R%x702%h", "#R#r");
	fl_addtopup(p[1], "Load\t\tAlt-L%x703%h", "#L#l");
	fl_addtopup(p[1], "Save\t\tAlt-S%x704%h", "#S#s");

	fl_addtopup(p[2], "Close\t\tAlt-O%x8%h", "#o#O");
	fl_addtopup(p[2], "Exit\t\tAlt-E%x9%h", "#E#e");
	fl_set_menu_popup(menu, p[2]);
	menu->u_ldata=(long)p;/* so as to free pup when destroying chnannel window l8r*/
}

/*----------------bookmark_menu-------------------------------------------------------*/

#if DEBUG>2
static void print_bm_tree(Bookmark * p, int pos)
{
	if (p) {
		print_bm_tree(p->right, pos + 8);
		printf("%*.10s %d\n", pos,
		       p->channel, p->bf);
		print_bm_tree(p->left, pos + 8);
	}
}
#endif

static int item2id(int item)
{
/* pupup menu: convert menu value to bm id */
	if (item < 30)
		item %= 20;
	else if (item < 300)
		item %= 200;
	else if (item < 3000)
		item %= 2000;
	else
		item %= 20000;	/* 29999%20000==9999 max entries !*/
	return item;
}

static void do_bookmark_menu(Winstruct * win)
{
	static void *find_bookmark(int id, Bookmark * bm);

	int     item;
	Bookmark *bm;

	/*bookmarks_setup(win->chanwin->bookmark_menu); uncomment */
#if DEBUG>4
	print_bm_tree(bmTree, 8);
#endif
	if ((item = fl_get_menu(win->chanwin->bookmark_menu)) <= 0)
		return;
	if (item == 1) {	/* Add bookmark */
		int     id;

		if (win->chanCur == win->chanStart)
			return;	/* no channels */
		bm = (Bookmark *) malloc(sizeof(Bookmark));
		if (!bm) {
			perror("malloc");
			return;
		}
		bm->channel = strdup(win->chanCur->name);
		if (win->chanCur->topic) {
			bm->title = strdup(win->chanCur->topic);
		}
		else
			bm->title = NULL;
		if (win->server) {
			bm->server = strdup(win->server->servname);
			bm->port = win->server->port;
		}
		else {
			bm->server = NULL;
			bm->port = DEFAULT_PORT;
		}
		bm->id=++bm_id;
		insert_bookmark(&bmTree, bm, &id);
		bookmarks_setup(win->chanwin->bookmark_menu);
		return;
	}
else{
	char *buf;

	item = item2id(item);
	bm = find_bookmark(item, bmTree);
	if (!bm) {
		fl_addto_browser(win->chanwin->brow2, PROMPT "bookmark bug: Please try again. /bugs for info ;)");
		return;
	}
	if ((gflags & BKMRK_IG_SERVER) || bm->server == NULL) {
		buf= (char *) alloca(sizeof(char) * (strlen(bm->channel)+8));
		sprintf(buf, "join %s", bm->channel);
		process_cmd(buf, win);
		return;
	}
	else {
		static Winstruct *get_server_win(char *name, int port);
		Winstruct *w = get_server_win(bm->server, bm->port);
		
		buf= (char *) alloca(sizeof(char) * (strlen(bm->channel)+8));
		sprintf(buf, "join %s", bm->channel);
		if (w && w->server->fd>-1) {
			/* already connected to server */
			process_cmd(buf, w);
			return;
		}
		else {		/* connect to server */
			
			Server  *s;
		   if(w) s=w->server;
		   else{
			s=(Server *)malloc(sizeof(Server));
			if(!s)
				error(-2, "malloc error");
			s->email = (char *)malloc(sizeof(char)* (strlen(Name) + strlen(host) + 3));
			s->ircname = (char *)malloc(sizeof(char)*(strlen(sula_NAME sula_VERSION) + 8));
			sprintf(s->email, "%s@%s", Name, host);
			s->ircname = sula_NAME sula_VERSION " user";	/* dont want to mess with pwd anymore.. */
			s->port = bm->port;
			s->servname = bm->server;
			s->password = NULL;
			s->nick = Name;
			s->name=NULL;
			s->mode=NULL;
			s->flag =0;
			s->read=0;
			s->lag=NULL;
			s->lastInvite=NULL;
			s->fd= -1;
			s->state=-1;
			s->iNick=-1;
		     }
			add_cmdoncon(CLI_CMD, buf, s->servname, s->port);
			connect_to_server(s, w?w:NULL,NULL,0);
		}
	}
 }/*else*/
}

static Winstruct *get_server_win(char *name, int port)
{
/* return first window on this server */
	register i;
	for(i=0;i<win_count;i++)
		if (winstruct[i].chanwin && winstruct[i].server && 
		!strcasecmp(name, winstruct[i].server->servname) &&
		    port == winstruct[i].server->port)
		    return &winstruct[i];
	return NULL;
}

static Bookmark *bm_copy;	/* copy of original bookmarks */

static void *find_bookmark(int id, Bookmark * bm)
{
	/*Bookmark *bm=bmTree; */
	while (bm) {
		if (id == bm->id)
			return bm;
		else if (id < bm->id)
			bm = bm->left;
		else
			bm = bm->right;
	}
	return NULL;
}

static void bookmarks_setup(FL_OBJECT * menu)
{
	int p;
	static void add_entry(int *popup, Bookmark * bm);
	if (menu->u_ldata!=-1)
		fl_freepup(menu->u_ldata);
retry:
	p = fl_newpup(FL_ObjWin(menu));
	if(p==-1){
		inc_pup_max();
		goto retry;
	}
	fl_setpup_shadow(p, 0);
	fl_setpup_cursor(p, XC_right_ptr);
	fl_setpup_bw(p, -1);
	fl_addtopup(p, "Add bookmark\tAlt-N%x1%h%l", "#A#a");
	add_entry(&p, bmTree);

	fl_set_menu_popup(menu, p);
	menu->u_ldata=p;
}

static void add_entry(int *pup, Bookmark * bm)
{

	if (bm) {
		char   *buf;
	       	add_entry(pup, bm->left);
		if (bm->title)
			buf = (char *)malloc(sizeof(char)* (strlen(bm->title) + 32));

		else
			buf = (char *)malloc(sizeof(char)* (strlen(bm->channel) + 32));
		if (!buf)
			return;
		sprintf(buf, "%d. %s%%x2%d", bm->id, bm->title ? bm->title : bm->channel, bm->id);
#if DEBUG>2
		fprintf(stderr, "%s\n", buf);
#endif
		fl_addtopup(*pup, buf);
		free(buf);

		add_entry(pup, bm->right);
	}
	return;
}

/*----------------------*/
/* window->bookmarks    */
/*----------------------*/

static void add_bm_to_brow(FL_OBJECT * brow, Bookmark * bm)
{
	if (bm) {
		char   *buf;

		add_bm_to_brow(brow, bm->left);
		if (bm->title)
			buf = (char *)malloc(sizeof(char)* (strlen(bm->title) + 32));

		else
			buf = (char *)malloc(sizeof(char)* (strlen(bm->channel) + 32));
		if (!buf) 
			return;
		sprintf(buf, "%3d. %s", bm->id, bm->title ? bm->title : bm->channel);
		fl_add_browser_line(brow, buf);
		free(buf);
		add_bm_to_brow(brow, bm->right);
	}

}

static int bm_form_atclose(FL_FORM * form, void *data)
/* don't use the value  of form. might be NULL .see further below*/
{
	FD_bookmarks **f = (FD_bookmarks **) data;

	if ((*f)->bar->u_vdata) {
		fl_hide_form((FL_FORM *) (*f)->bar->u_vdata);
		fl_free_form((FL_FORM *) (*f)->bar->u_vdata);
	}
	fl_freepup( *((int *)	 (*f)->file->u_ldata));
	fl_freepup( * ( (int *)	 (*f)->file->u_ldata +1));
	if(!form){
		/* menu File->Close */
		fl_hide_form((*f)->bookmarks);
		fl_free_form((*f)->bookmarks);
		*((FD_bookmarks **) (*f)->file->u_vdata) = NULL;
	}else  /* win manager*/
	*f = NULL;
	return FL_OK;
}

static Bookmark *copy_bm(Bookmark * bm)
{
/* make a copy of the bm tree */
	Bookmark *tmp;

	if (bm) {
		tmp = (Bookmark *) malloc(sizeof(Bookmark));
		if (!tmp)
			return NULL;
		tmp->left = copy_bm(bm->left);
		tmp->right = copy_bm(bm->right);
		tmp->channel = strdup(bm->channel);
		tmp->server = strdup2(bm->server);
		tmp->port = bm->port;
		tmp->title = strdup2(bm->title);
		tmp->id = bm->id;
		return tmp;
	}
	return NULL;
}

static int bm_equal(Bookmark * a, Bookmark * b)
{

	return ((!a && !b) ||
		(a && b &&
		 (/*a->id == b->id &&*/ a->port == b->port &&
		 	!strcasecmp(a->channel, b->channel) &&
		   !strcasecmp(a->server, b->server)) &&
		 bm_equal(a->left, b->left) &&
		 bm_equal(a->right, b->right))
	   );
}

static void do_bookmark_win(Winstruct * win)
{
	int     p[2]; /* pup ID's */
	static FD_bookmarks *bm_form = NULL;
	static void item_menu_setup(FL_OBJECT * menu, int browlines, int *pupup);

	if (bm_form) {
		fl_raise_form(bm_form->bookmarks);
		return;
	}
	bm_copy = copy_bm(bmTree);
	if (!bm_equal(bm_copy, bmTree)) {
		error(0, "[%s %d] error copying bookmarks.Using original..\n",
		 __FILE__, __LINE__);
		bm_copy = bmTree;
	}
	bm_form = create_form_bookmarks();
	fl_set_form_atclose(bm_form->bookmarks, bm_form_atclose, (void *) &bm_form);
	bm_form->file->u_vdata = (void *) &bm_form;
	bm_form->file->u_ldata =(long) p;
	fl_set_browser_fontsize(bm_form->browser, FL_NORMAL_SIZE);

	fl_set_object_bw(bm_form->bar, -2);
	bm_form->bar->u_vdata = (void *) 0;	/* no sub windows */
	add_bm_to_brow(bm_form->browser, bm_copy);
	fl_set_form_minsize(bm_form->bookmarks, 250, 200);
	
	p[0] = fl_newpup(0);
	fl_setpup_shadow(p[0], 0);
	fl_setpup_cursor(p[0], XC_right_ptr);
	fl_setpup_bw(p[0], -1);
	fl_addtopup(p[0], "Open%x1%i");
	fl_addtopup(p[0], "Import%x2");
	fl_addtopup(p[0], "Save%x3%l");
	fl_addtopup(p[0], "Close%x4");
	fl_set_menu_popup(bm_form->file, p[0]);
	item_menu_setup(bm_form->item, 0, &(p[1]));

	fl_show_form(bm_form->bookmarks, FL_PLACE_FREE | FL_FREE_SIZE, FL_TRANSIENT,
		     sula_NAME sula_VERSION ": Bookmark editor");
	fl_do_forms();
}

static void do_bm_file_menu(FD_bookmarks * w)
{
	int     item = fl_get_menu(w->file);

	if (item < 1)
		return;
	switch (item) {
	    case 1:
		    break;	/* open */
	    case 2:{		/* import */
			    char   *fname;
			    Bookmark *tmp = NULL;

			    fl_set_cursor(w->bookmarks->window, XC_watch);
			    fl_deactivate_form(w->bookmarks);
			    fname = getstr2("File name:", sula_NAME sula_VERSION ": Load bookmarks", NULL, 1);
			    if (!fname){
			    	fl_activate_form(w->bookmarks);
			    	fl_reset_cursor(w->bookmarks->window);
			    	return;
			    }	
			    if (load_bookmarks(fname, item == 1 ? &tmp : &bm_copy)) {
				    char   *buf = (char *)malloc(sizeof(char)*(strlen(fname) + 128));
				    sprintf(buf, "%s\n%s", fname, strerror(errno));
				    confirm(buf, sula_NAME sula_VERSION ": Error", 0);
				    free(buf);
			    }
			    else {
				    if (item == 1)
					    bm_copy = copy_bm(tmp);
				    fl_clear_browser(w->browser);
				    add_bm_to_brow(w->browser, bm_copy);
			    }
			    fl_activate_form(w->bookmarks);
			    fl_reset_cursor(w->bookmarks->window);
			    break;
		    }
	    case 3:{		/* save */
			    char   *fname;

			    fl_set_cursor(w->bookmarks->window, XC_watch);
			    fl_deactivate_form(w->bookmarks);
			    fname = getstr2("Save as:",
					    sula_NAME sula_VERSION ": Save bookmarks", NULL, 1);
			    if (!fname){
			    	fl_activate_form(w->bookmarks);
			    	fl_reset_cursor(w->bookmarks->window);
			    	return;
			    }	
			    if (save_bookmarks(fname, bm_copy))
				    error(-1,"save bookmarks:%s", fname);
			    fl_activate_form(w->bookmarks);
			    fl_reset_cursor(w->bookmarks->window);
			    break;
		    }
	    case 4:		/* close */
			    
#if 1
		if (!bm_equal(bm_copy, bmTree) &&
			confirm("Bookmarks have changed. Use changes?",
	 		sula_NAME sula_VERSION" Confirm bookmark changes", 10))
		{
			int i;
			bmTree = copy_bm(bm_copy);
			if (!bm_equal(bm_copy, bmTree))
			  fprintf(stderr, "error copying bookmarks\n");
			#if 1
			else
			for(i=0;i<win_count;i++)
		    	  if(winstruct[i].chanwin){
			     fl_freepup(winstruct[i].chanwin->bookmark_menu->u_ldata);
		    	     bookmarks_setup(winstruct[i].chanwin->bookmark_menu);
			 }
			#endif
		}
#endif
		bm_form_atclose(NULL, &w);
		break;
	    default:
		    break;
	}
}
static void item_menu_setup(FL_OBJECT * menu, int count, int *p)
/* count:no of bookmarks*/
{
retry:
	*p = fl_newpup(FL_ObjWin(menu));
	if(*p==-1){
		inc_pup_max();
		goto retry;
	}
	fl_setpup_shadow(*p, 0);
	fl_setpup_cursor(*p, XC_right_ptr);
	fl_setpup_bw(*p, -1);
		fl_addtopup(*p, "Properties%x1");
		fl_addtopup(*p, "Go to bookmark%x2%i");
		fl_addtopup(*p, "Delete%x3");
		fl_addtopup(*p, "New%x4");
	fl_set_menu_popup(menu, *p);
}

static void do_bm_item_menu(FD_bookmarks * w)
{
	int     item, n;

	item = fl_get_menu(w->item);
	if (item < 1)
		return;
	n = fl_get_browser(w->browser);
	if (n < 1 && item != 4)
		return;

	switch (item) {
	    case 4:		/* new */
	    case 1:{		/* properties */

			    FD_edit_bm *form;
			    Bookmark *bm;
			    char    pp[8];

			    if (item != 4) {
				    int     id;
				    char   *s = (char *) fl_get_browser_line(w->browser, n);

				    if (!s || !(*s))
					    return;
				    if (sscanf(s, "%d.", &id) < 1)
					    return;
				    bm = find_bookmark(id, bm_copy);
				    if (!bm)
					    return;
			    }
			    else
				    bm = NULL;

			    fl_set_cursor(w->bookmarks->window, XC_watch);
			    fl_deactivate_form(w->bookmarks);

			    form = create_form_edit_bm();
			    w->bar->u_vdata = (void *) form->edit_bm;
			    form->channel->u_ldata = form->server->u_ldata = form->title->u_ldata =
			       form->port->u_ldata = (long) XC_xterm;
			    fl_set_object_prehandler(form->channel, set_cursor);
			    fl_set_object_prehandler(form->server, set_cursor);
			    fl_set_object_prehandler(form->title, set_cursor);
			    fl_set_object_prehandler(form->port, set_cursor);

			    if (item != 4) {
				    fl_set_input(form->channel, bm->channel);
				    fl_set_input(form->server, bm->server ? bm->server : "any");
				    if (bm->title)
					    fl_set_input(form->title, bm->title);
				    sprintf(pp, "%d", bm->port);
				    fl_set_input(form->port, pp);
				    form->ok->u_ldata = (long) &n;	/* selectedline */
				    form->add->u_vdata = (void *) bm;	/* selected bm */
			    }
			    else {
				    fl_hide_object(form->ok);
				    fl_set_input(form->server, "any");
				    sprintf(pp, "%d", DEFAULT_PORT);
				    fl_set_input(form->port, pp);
			    }
			    form->cancel->u_vdata = (void *) w;
			    fl_set_form_atclose(form->edit_bm, dont_close_win, 0);
			    fl_set_form_minsize(form->edit_bm, 410, 210);
			    fl_set_form_maxsize(form->edit_bm, 610, 210);
			    fl_show_form(form->edit_bm, FL_PLACE_FREE | FL_FREE_SIZE, FL_TRANSIENT, sula_NAME sula_VERSION ": Edit bookmark");
			    fl_do_forms();
			    break;
		    }
	    case 2:{		/* go to */
			    break;
		    }
	    case 3:{		/*delete */
			    int     id;
			    Bookmark *bm;
			    char   *s = (char *) fl_get_browser_line(w->browser, n);

			    if (!(*s))
				    return;
			    if (sscanf(s, "%d.", &id) < 1)
				    return;
			    bm = find_bookmark(id, bm_copy);
			    if (!bm)
				    return;
			    del_bookmark(&bm_copy, bm);
			    fl_freeze_form(w->bookmarks);
			    fl_clear_browser(w->browser);
			    add_bm_to_brow(w->browser, bm_copy);
			    fl_unfreeze_form(w->bookmarks);
			    break;
		    }
	    default:
		    break;
	}
}

void    edit_bm_cb(FL_OBJECT * ob, long data)
{
	FD_edit_bm *form = (FD_edit_bm *) ob->form->fdui;	/* editing form */
	FD_bookmarks *f = (FD_bookmarks *) form->cancel->u_vdata;	/* bookmark form with browser */

	switch (data) {

	    case 0:		/* ok; replace */
	    case 1:{		/*add */

			    char   *port = NULL;
			    int     id;		/* of new bookmark */
			    Bookmark *bm, *new_bm;

			    bm = (Bookmark *) malloc(sizeof(Bookmark));
			    if (!bm) {
				    error(-1,"malloc error");
				    return;
			    }
			    bm->channel = nextstr((char *)fl_get_input(form->channel), 1);
			    if (*bm->channel==0) {
				    confirm("NO channel name given", sula_NAME sula_VERSION ": Error", 0);
				    return;
			    }
			    bm->server = nextstr((char *)fl_get_input(form->server), 1);
			    if (*bm->server && !strcasecmp(bm->server, "any"))
				    bm->server = NULL;
			    bm->title = strdup2((char *)fl_get_input(form->title));
			    port = nextstr((char *)fl_get_input(form->port), 1);
			    if (*port==0)
				    bm->port = DEFAULT_PORT;
			    else {
				    int     i = atoi(port);
				    size_t  ll;

				    if (i < 0)
					    *port = '\0';
				    ll = strlen(port);
				    while ((i = atoi(port)) > USHRT_MAX)
					    *(port + --ll) = '\0';
				    if ((bm->port = i) == 0)
					    bm->port = DEFAULT_PORT;
				  free(port);
			    }
			    if (!data) {	/* replace */
				    Bookmark *old = (Bookmark *) form->add->u_vdata;
				    bm->id=old->id;
				    del_bookmark(&bm_copy, old);
			    }
			    else bm->id=++bm_id;
			    insert_bookmark(&bm_copy, bm, &id);
			    new_bm = find_bookmark(bm->id, bm_copy);
			    #if 1
			    assert(new_bm);
			    #else
			    if (new_bm == NULL) {
				    confirm("Bookmark error. Might have lost 1",
				     sula_NAME sula_VERSION ": Error", 0);
				    break;	/* should quit rreally prog now */
			    }
			    #endif
			    free(bm->channel);
			    free(bm->server);
			    free(bm->title);
			    fl_clear_browser(f->browser);
			    fl_freeze_form(f->bookmarks);
			    add_bm_to_brow(f->browser, bm_copy);
			    fl_unfreeze_form(f->bookmarks);
			    break;
		    }
	    case 2:		/* CANCEL */
		    break;
	    default:
		    break;
	}
	fl_reset_cursor(f->bookmarks->window);
	fl_activate_form(f->bookmarks);
	f->bar->u_vdata = (void *) 0;	/*we no longer exist */
	fl_hide_form(form->edit_bm);
	fl_free_form(form->edit_bm);
}
void    bkmark_cb(FL_OBJECT * ob, long data)
{
	switch (data) {
	    case 0:		/*file */
		    do_bm_file_menu((FD_bookmarks *) ob->form->fdui);
		    break;
	    case 1:		/* item */
		    do_bm_item_menu((FD_bookmarks *) ob->form->fdui);
		    break;
	    default:
		    break;
	}
}

extern Winstruct *new_win(void)
{
	Winstruct *win= new_chan_win();
	if (win)
		set_object_attribs(win);
	return win;
}
static int win_alloc(){
	int i;
	if(winstruct==NULL){
		winstruct=malloc(sizeof(Winstruct)*WIN_ALLOC);
	}
	else{
		winstruct=realloc(winstruct, (win_count+WIN_ALLOC)*sizeof(Winstruct));
	}
	if(winstruct==NULL) return (-1);
	for (i=win_count;i< win_count+WIN_ALLOC; i++){
		winstruct[i].chanwin=NULL;
		winstruct[i].server=NULL;	/* new_process_cmd needs this*/

	}
	win_count+=WIN_ALLOC ;
	return (0);
}
static int win_raw_cb(FL_FORM *form, void *xevent){
	
	switch(((XEvent *)xevent)->type){
	    case KeyPress:
		  switch(XLookupKeysym((XKeyEvent *)xevent,0)){
		    case XK_Tab:
			{
		/* use TAB key to insert nicks into input field */
			fl_trigger_object(((FD_chanwin *)form->fdui)->tab);
			return FL_PREEMPT;
			}
		    case XK_Escape:
		  /* use ESC key to show/hide control window */
		    {
		  	static FL_Coord x=0, y=0;
			if(gflags&CTRL_WIN_SHOWN){
				x=main_form->main_form->x;
				y=main_form->main_form->y;
				fl_hide_form(main_form->main_form );
				gflags &= ~CTRL_WIN_SHOWN;
			}else
			{
			 	fl_set_form_position(main_form->main_form,x,y);
				fl_show_form(main_form->main_form,
				FL_PLACE_POSITION, FL_FULLBORDER, 
				sula_NAME sula_VERSION ": Control");

				gflags |= CTRL_WIN_SHOWN;
			}
			return FL_PREEMPT;
		    }
		   case XK_C:
		   /* Ctrl-C to cycle channels anti-clockwise */
		   	if(((XKeyEvent *)xevent)->state & ControlMask){
				fl_trigger_object(((FD_chanwin *)form->fdui)->chan_left);
				return FL_PREEMPT;
			}
		   case XK_c:
		/* Ctrl-c to cycle channels clockwise */
			if(((XKeyEvent *)xevent)->state & ControlMask){
				fl_trigger_object(((FD_chanwin *)form->fdui)->chan_right);
				return FL_PREEMPT;
			}
		}/*switch (XLookupKeysym(*/
	   default: break;
	}
	return !FL_PREEMPT;
}


extern Winstruct *new_chan_win(void){
	int i;
	if(winstruct==NULL && win_alloc()) 
		return (NULL);
retry:
	for(i=0;i< win_count; i++)
		if(winstruct[i].chanwin==NULL){
		Winstruct *win;
	Go     *p;
	win=& winstruct[i];
	win->chanwin= create_form_chanwin();
	fl_set_form_atclose(win->chanwin->chanwin, dont_close_win, 0);
	win->chanwin->bar->u_ldata = (long)i;		/* this window; used in posthandler etc*/
	/*win->chanwin->clock->u_ldata = (long) 0;*/	/* no go pup */
	win->chanwin->status->u_ldata = (long) 0;	/* no of entries in Go_menu */
	win->chanwin->chanusers->u_vdata = NULL;	/* chanusers not open*/
	win->chanwin->find->u_vdata = NULL;		/* find not open*/
	win->chanwin->minput->u_vdata = NULL;

	/*fl_set_object_prehandler(win->chanwin->sinput, set_cursor);*/
	win->chanwin->sinput->u_ldata = (long) XC_xterm;
	fl_set_object_prehandler(win->chanwin->sinput, set_cursor);

	win->chanwin->browser->u_ldata = (long) XC_right_ptr;
	fl_set_object_prehandler(win->chanwin->browser, set_cursor);
	
	win->server = NULL;
	win->chanwin->cmd_fwd->u_ldata = (long) histCur;
	win->chanwin->tab->u_ldata=(long) senderCur;/* TAB key */
	win->chanStart = (Channel *) malloc(sizeof(Channel));
	if (win->chanStart == NULL)
		goto errur;
	win->chanStart->left = win->chanStart->right = win->chanCur = win->chanStart;
	win->dccStart = (DCCChat *) malloc(sizeof(DCCChat));
	if (win->dccStart == NULL)
		goto errur;
	win->dccStart->left = win->dccStart->right = win->dccCur = win->dccStart;
	win->lastJoin = (NoChannel *) malloc(sizeof(NoChannel));
	win->lastPart = (NoChannel *) malloc(sizeof(NoChannel));
	if (!win->lastJoin || !win->lastPart)
		goto errur;
	win->lastJoin->name = NULL;
	win->lastJoin->key = NULL;
	win->lastPart->name = NULL;
	win->lastPart->key = NULL;
	win->goStart = win->goEnd = (Go *) malloc(sizeof(Go));
	if (win->goStart == NULL)
		goto errur;
	p = win->goStart;
	win->goStart = (Go *) malloc(sizeof(Go));
	if (win->goStart == NULL)
		goto errur;
	win->goStart->name = strdup("#sula");	/* okay.otherwise free() gets stuck */
	win->goStart->key = NULL;
	win->goStart->next = p;
	win->query_nick = NULL;
	win->chanwin->go_menu->u_ldata=-1;
	win->chanwin->file_menu->u_ldata=-1;
	win->chanwin->go_menu->u_ldata=-1;
	win->chanwin->bookmark_menu->u_ldata=-1;
	win->chanwin->edit_menu->u_ldata=-1;
	win->chanwin->window_menu->u_ldata=-1;
	win->chanwin->help_menu->u_ldata=-1;
	menu_setup(win);
	go_menu_setup(win);

	fl_register_raw_callback(
		win->chanwin->chanwin,KeyPressMask, win_raw_cb);

	return (&winstruct[i]);
errur:
	perror("Error");
	winstruct[i].chanwin=NULL;
	return (NULL);
	}
      if(win_alloc()) return (NULL);
	goto retry;

}


static int win_ph(FL_OBJECT *, int, FL_Coord, FL_Coord, int, void *);

extern void set_object_attribs(Winstruct * win)
{
#include "closed.xpm"
#include "down_arrow.xbm"
#include "up_arrow.xbm"

#include "join.xpm"
#include "part.xpm"
#include "dcc.xpm"
#include "chanusers.xpm"
#include "find.xpm"
#include "minput.xpm"
#include "copy.xpm"
#include "cut.xpm"
#include "paste.xpm"
#include "delete.xpm"
#ifdef 0
Pixmap P, mask;
unsigned width,height;
#include  "win.xpm"

	 P=fl_create_from_pixmapdata(fl_default_window(),
	  win_xpm, &width,&height, &mask,0,0,0);
	fl_set_form_icon(win->chanwin->chanwin, P, mask);
#endif
	fl_set_object_bw(win->chanwin->browser, -2);
	fl_set_object_bw(win->chanwin->bar, -2);
	fl_set_object_bw(win->chanwin->msg_radio_grp, 2);

	fl_set_browser_fontsize(win->chanwin->browser, browser_fontsize);
	fl_set_browser_fontstyle(win->chanwin->browser, browser_fontsyle);
	fl_set_browser_fontsize(win->chanwin->brow2, browser2_fontsize);
	fl_set_browser_fontstyle(win->chanwin->brow2, browser2_fontsyle);
	fl_set_object_color(win->chanwin->browser,browser_colour, 
	browser_selected_colour);
	fl_set_object_color(win->chanwin->brow2,browser2_colour, 
	browser2_selected_colour);
	fl_set_object_lcolor(win->chanwin->browser,browser_textcolour);
	fl_set_object_lcolor(win->chanwin->brow2,browser2_textcolour);
	fl_set_object_lstyle(win->chanwin->sinput, input_fontstyle);
	fl_set_input_color(win->chanwin->sinput, input_colour,FL_BLUE);
	fl_set_choice_fontstyle(win->chanwin->target, target_fontsyle);
	fl_set_choice_fontsize(win->chanwin->target, target_fontsize);
	fl_set_object_color(win->chanwin->target, target_colour,target_colour2);

	fl_set_input_return(win->chanwin->sinput, FL_RETURN_END);
	fl_set_choice_align(win->chanwin->target, FL_ALIGN_LEFT);
	#ifdef _GUILE
	fl_set_object_label(win->chanwin->r_quote,"Parse F1");
	#endif
	fl_set_button(win->chanwin->r_quote, 1);
	fl_set_button(win->chanwin->type, 0);

	fl_set_button(win->chanwin->history, sflags & LOG);
	fl_hide_object(win->chanwin->info);
	fl_hide_object(win->chanwin->away);

	fl_set_pixmapbutton_data(win->chanwin->close_con, closed_xpm);
	fl_set_pixmapbutton_data(win->chanwin->join, join_xpm);
	fl_set_pixmapbutton_data(win->chanwin->part, part_xpm);
	fl_set_pixmapbutton_data(win->chanwin->dcc, dcc_xpm);
	fl_set_pixmapbutton_data(win->chanwin->chanusers, chanusers_xpm);
	fl_set_pixmapbutton_data(win->chanwin->minput, minput_xpm);
	fl_set_pixmapbutton_data(win->chanwin->find, find_xpm);
	fl_set_pixmapbutton_data(win->chanwin->copy, copy_xpm);
	fl_set_pixmapbutton_data(win->chanwin->cut, cut_xpm);
	fl_set_pixmapbutton_data(win->chanwin->paste, paste_xpm);
	fl_set_pixmapbutton_data(win->chanwin->remove, delete_xpm);
	fl_set_bitmapbutton_data(win->chanwin->chan_right, up_arrow_width,
				 up_arrow_height, up_arrow_bits);
	fl_set_bitmapbutton_data(win->chanwin->chan_left, dwn_arrow_width,
				 dwn_arrow_height, dwn_arrow_bits);

	
	
	fl_addto_choice(win->chanwin->operation,
	  "msg|notice|action|whois|chat|clientinfo|finger|ping|time|version"
	      "|userinfo|chelp|kick *|op|deop|invite|userhost|ison|whowas|query");
	fl_addto_choice(win->chanwin->action, "notice|action|msg");
	fl_set_form_minsize(win->chanwin->chanwin, 515, 560);
	fl_set_form_maxsize(win->chanwin->chanwin, 632, 560);
	/*fl_set_object_posthandler(win->chanwin->sinput, win_ph);*/
	fl_set_object_posthandler(win->chanwin->join, win_ph);
	fl_set_object_posthandler(win->chanwin->part, win_ph);
	fl_set_object_posthandler(win->chanwin->dcc, win_ph);
	fl_set_object_posthandler(win->chanwin->chanusers, win_ph);
	fl_set_object_posthandler(win->chanwin->minput, win_ph);
	fl_set_object_posthandler(win->chanwin->paste, win_ph);
	fl_set_object_posthandler(win->chanwin->copy, win_ph);
	fl_set_object_posthandler(win->chanwin->cut, win_ph);
	fl_set_object_posthandler(win->chanwin->remove, win_ph);
	fl_set_object_posthandler(win->chanwin->lag, win_ph);
	fl_set_object_posthandler(win->chanwin->away, win_ph);
	fl_set_object_posthandler(win->chanwin->find, win_ph);
	fl_set_object_posthandler(win->chanwin->info, win_ph);
	fl_set_object_posthandler(win->chanwin->close_con, win_ph);
	fl_set_object_posthandler(win->chanwin->chan_left, win_ph);
	fl_set_object_posthandler(win->chanwin->chan_right, win_ph);
	fl_set_object_posthandler(win->chanwin->topic, win_ph);
	fl_set_object_posthandler(win->chanwin->r_quote, win_ph);
	fl_set_object_posthandler(win->chanwin->r_query, win_ph);
	fl_set_object_posthandler(win->chanwin->r_channel, win_ph);
	fl_set_object_posthandler(win->chanwin->r_dcc_chat, win_ph);
	fl_set_object_posthandler(win->chanwin->r_echo, win_ph);
	fl_set_object_posthandler(win->chanwin->type, win_ph);
	fl_set_object_posthandler(win->chanwin->action, win_ph);
	fl_set_object_posthandler(win->chanwin->go, win_ph);
	fl_set_object_posthandler(win->chanwin->nick, win_ph);
	fl_set_object_posthandler(win->chanwin->history, win_ph);
	fl_set_object_posthandler(win->chanwin->do_it, win_ph);
	fl_set_object_posthandler(win->chanwin->operation, win_ph);
	fl_set_object_posthandler(win->chanwin->update, win_ph);

	fl_trigger_object(win->chanwin->update);
}

static Ph ph;
void do_tips(int id, void *data)
{
        Ph     *p = data;

        fl_show_oneliner(p->s, p->ob->form->x + p->ob->x,
                         p->ob->form->y + p->ob->y + p->ob->h + 2);
        timeoutID = fl_add_timeout((double) tips_tmout, do_tips, (void *) p);
}

static int win_ph(FL_OBJECT * ob, int event, FL_Coord mx, FL_Coord my, int key, void *xev)
{

	if (!(gflags & SHOW_TIPS))
		return 0;
	else{
	FD_chanwin *chanwin=(FD_chanwin *)ob->form->fdui;
	ph.ob = ob;
	ph.s = NULL;
	if (ob == chanwin->join)
		ph.s = "join new|last|invite";
	else if (ob == chanwin->part)
		ph.s = "part current|all|last";
	else if (ob == chanwin->dcc)
		ph.s = "chat accept|in new win|close";
	else if (ob == chanwin->chanusers)
		ph.s = "channel members|list|update";
	else if (ob == chanwin->away){
		Winstruct *win;
		win = &winstruct[chanwin->bar->u_ldata];
		ph.s = (win->server && win->server->flag & SET_AWAY) ? 
		"mark present" : "mark away";
	}
	else if (ob == chanwin->minput)
		ph.s = "compose multiline message";
	else if (ob == chanwin->copy)
		ph.s = "copy from 1. | 2. browser";
	else if (ob == chanwin->cut)
		ph.s = "cut from 1. | 2. browser";
	else if (ob == chanwin->remove)
		ph.s = "delete from 1. | 2. browser";
	else if (ob == chanwin->paste)
		ph.s = "paste into input|browsers";

	else if (ob == chanwin->find)
		ph.s = "find line with text";
	else if (ob == chanwin->info)
		ph.s = "not in use..";
	else if (ob == chanwin->close_con) {
		int i=chanwin->bar->u_ldata;
		Winstruct *win;
		assert(!win_invalid(i));
		win = &winstruct[i];
		if (win->server && win->server->fd > -1)
			ph.s = "disconnect from server";
		else
			ph.s = "reconnect to server";
	}
	else if (ob == chanwin->lag)
		ph.s= "force lag update";
	else if (ob == chanwin->chan_left)
		ph.s = "<< channel | chat";
	else if (ob == chanwin->chan_right)
		ph.s = ">> channel| chat";
	else if (ob == chanwin->topic)
		ph.s = "change|clear|query topic";
	else if (ob == chanwin->r_quote)
		#ifdef _GUILE
		ph.s = "execute input as Scheme code";
		#else
		ph.s = "quote input to server";
		#endif
	else if (ob == chanwin->r_query)
		ph.s = "feed input to query";
	else if (ob == chanwin->r_channel)
		ph.s = "feed input to channel";
	else if (ob == chanwin->r_dcc_chat)
		ph.s = "feed input to DCC chat";
	else if (ob == chanwin->r_echo)
		ph.s = "echo input";
	else if (ob == chanwin->type)
		ph.s = fl_get_button(ob) ? "deactivate deflt msg type":"use default msg type" ;
	else if (ob == chanwin->action)
		ph.s = "default message type";
	else if (ob == chanwin->history)
		ph.s = fl_get_button(ob) ? "stop logging" : 
		"resume logging";
	else if (ob == chanwin->go)
		ph.s = "execute op using this msg";
	else if (ob == chanwin->nick)
		ph.s = "set nick|mode|channel mode";
	else if (ob == chanwin->do_it)
		ph.s = "<- execute op";
	else if (ob == chanwin->operation)
		ph.s = "operation";
	else if (ob == chanwin->update)
		ph.s = "update targets ->";
	else
	return 0;
	if (event == FL_ENTER)
                	timeoutID = fl_add_timeout((double) tips_tmout,
                     	do_tips, (void *) &ph);
        else if (event == FL_LEAVE || event == FL_PUSH) {
                	fl_hide_oneliner();
                	if (timeoutID) {
                         	fl_remove_timeout(timeoutID);
                         	timeoutID = 0;
                 	}
         }
        }
	return 0;
}

