/*
   D C C  & C T C P
 */

/* ctcp.c       - CTCP and some DCC protocol
   see also: dcc.c

   Author: Tano Fotang, 1997, 1998
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License , or
   (at your option) any later version.
   see the file COPYING for more info.
   -------------------------------------------------------
   ctcp_quote_it(),
   ctcp_unquote_it(),
   decrypt() and encrypt() are from the IrcII source code ,
   Written By Michael Sandrof and Copyright(c) 1990. GNU GPL.
   slighty modified.
   ----------------------------------------------------------

 */

#include <forms.h>
#if 1
#include <sys/utsname.h>	/* #define's UT_NAMESIZE etc for linux */
#else
#include <sys/types.h>		/* others use this...? */
#endif
#include <pwd.h>
#include <utmp.h>
#include <netdb.h>
#include <arpa/inet.h>
#include "config.h"
#include "aux.h"
#include "sula.h"
#include "chanwin.h"
#include "script.h"
#include "global.h"
#include "parser.h"
#include "dcc.h"
#include "setting.h"
#include "ignore.h"
#include "cmd.h"
#include "scheme.h"
#include "hooks.h"
#include "flood.h"

extern Winstruct *Win;		/* channel window */

/************************************************/
/*      c t c p                                 */
/************************************************/

/* taken f r o m   ircII  for compatibility */
#define CTCP_QUOTE_CHAR '\\'
#define CTCP_DELIM_CHAR '\001'
void    ctcp_unquote_it(char *, int *len);
char   *ctcp_quote_it(char *str, int len);
static void decrypt(char *, int, char *);

/* ---------------------------------------- */

static void addkey(Winstruct * win, Key ** s, char *nick, char *key)
{
	Key    *p = *s;
	char   *buf = NULL;

	while (p)
		if (!strcasecmp(p->who, nick)) {
			free(p->key);
			p->key = strdup(key);
			p->flag = 1;
			buf = malloc(sizeof(char) * (strlen(p->key) + strlen(p->who) + 30));

			sprintf(buf, "@ikey for %s changed to \"%s\"", p->who, p->key);
			break;
		}
		else
			p = p->next;
	if (p)
		return;
	else {

		p = *s;
		*s = (Key *) malloc(sizeof(Key));
		(*s)->next = p;
		(*s)->who = strdup(nick);
		(*s)->key = strdup(key);
		(*s)->flag = 1;
		buf = malloc(sizeof(char) * (strlen((*s)->key) + strlen((*s)->who) + 40));

		sprintf(buf, "@i%s added to crypt list with key \"%s\"", (*s)->who, (*s)->key);
	}
	if (buf) {
		say(buf, win - winstruct, 1);
		free(buf);
	}
}
static void del_key_node(Key ** head, Key * trail, Key * node)
{
	if (trail)
		trail->next = node->next;
	else {
#warning uncomment and  test this now!!!!
		/*free((*head)->nick);free((*head)->key); */
		*head = (*head)->next;
		free(node);
	}
}
static void remove_key(Winstruct * win, char *who)
{
	Key    *p = keys;

	if (p) {
		char    found = 0;

		if (!strcasecmp(p->who, who)) {
			keys = keys->next;
			free(p);
			found++;
		}
		else {
			Key    *q = p->next;

			while (q)
				if (!strcasecmp(q->who, who)) {
					del_key_node(&(keys), p, q);
					found++;
					break;
				}
				else {
					p = q;
					q = q->next;
				}
		}
		if (!found)
			return;
		say2(win - winstruct, 1, "@i%s removed from crypt list", who);
	}
}
static void toggle_sed_flag(Winstruct * win, char *who)
{
	Key    *p = keys;

	while (p)
		if (!strcasecmp(p->who, who)) {
			p->flag = p->flag ? 0 : 1;
			say2(win - winstruct, 1, "@iencryption to %s now set %s",
			     p->who, p->flag ? "on" : "off");
			break;
		}
		else
			p = p->next;

}

char   *have_key(char *who, int action)
{
	Key    *p = keys;

	while (p)
		if (!strcasecmp(p->who, who)) {
			if (action == ENCRYPT && p->flag == 0)
				return NULL;
			else
				return p->key;
		}
		else
			p = p->next;
	return NULL;
}

extern void do_encrypt(int argc, char **argv, Winstruct * win)
{
/*
   kludge!!
 */

	if (argc == 1) {
		char    buf[256];
		Key    *p = keys;

		if (p) {
			sprintf(buf, "@b@f  %-20s%-20s  %s", "User/Channel", "Key", "Status");
			fl_addto_browser(win->chanwin->brow2, buf);
		}
		while (p) {
			sprintf(buf, "@f  %-20s%-20s  %s", p->who, p->key,
				p->flag ? "on" : "off");
			p = p->next;
			fl_addto_browser(win->chanwin->brow2, buf);
		}
		return;
	}
	if (!strncasecmp(argv[1], "-help", 2)) {
		show_std_help(uppercase(argv[0]), NULL, win->chanwin);
		return;
	}
	if (argc > 1)
		if (!strncasecmp(argv[1], "-del", 4) && argv[2])
			remove_key(win, argv[2]);
		else if (argc == 2)
			toggle_sed_flag(win, argv[1]);
		else if (argc > 2)
			addkey(win, &keys, argv[1], argv[2]);
}

static char *
        ctcp_unquote_it2(str, len)
/*  ctcp_unquote_it() breaks SED. Hence, this version, just for SED */
     char   *str;
     int    *len;
{
	char   *buffer;
	char   *ptr;
	char    c;
	int     i, new_size = 0;

	buffer = (char *) malloc(sizeof(char) * *len);

	if (!buffer)
		return NULL;
	ptr = buffer;
	i = 0;
	while (i < *len) {
		if ((c = str[i++]) == CTCP_QUOTE_CHAR) {
			switch (c = str[i++]) {
			    case CTCP_QUOTE_CHAR:
				    *(ptr++) = CTCP_QUOTE_CHAR;
				    break;
			    case 'a':
				    *(ptr++) = CTCP_DELIM_CHAR;
				    break;
			    case 'n':
				    *(ptr++) = '\n';
				    break;
			    case 'r':
				    *(ptr++) = '\r';
				    break;
			    case '0':
				    *(ptr++) = '\0';
				    break;
			    default:
				    *(ptr++) = c;
				    break;
			}
		}
		else
			*(ptr++) = c;
		new_size++;
	}
	*ptr = '\0';
	*len = new_size;
	return (buffer);
}

static void do_sed(void *junk)
{
/*
   this proc has very dirty code. i just want to get SED working now
 */
	#if CTCP_more
	if (!check_msg_hook(CTCP_SED, Win - winstruct, "%s,%d %s",
			    Win->server->servname, Win->server->port, Msg))
	#endif
	 {
		char   *buf;
		char   *nick, *dest, *login, *host, *msg, *sender;
		size_t  ll, ll2;
		char   *key = 0;
		char   *ptr = 0;

		ll = strlen(Msg);

		msg = (char *) malloc(sizeof(char) * ll);
		dest = (char *) alloca(sizeof(char) * ll);

		if (sscanf(Msg, "%*s PRIVMSG %s :\x1%[^\x1\n]\x1\n", dest, msg) < 2) {
			if (gflags & VERBOSE_CTCP)
				fprintf(stderr, "invalid CTCP sed: %s", Msg);
			return;
		}
		sender = nextstr(Msg, 1);
		if (*sender == '~' && *(sender + 1))
			strshift(sender,1);
		ll2 = strlen(sender);
		nick = (char *) alloca(sizeof(char) * ll2);
		login = (char *) alloca(sizeof(char) * ll2);
		host = (char *) alloca(sizeof(char) * ll2);

		if (sscanf(sender, "%[^!]!%[^@]@%s", nick, login, host) < 3)
			{free(sender);return;}
		key = have_key((*dest == '#' || *dest == '&') ? dest : nick, DECRYPT);
		if (key) {
			size_t  n;

			ptr = msg + 4;
			n = strlen(ptr);
			buf = ctcp_unquote_it2(ptr, &n);
			decrypt(buf, n, key);
			ctcp_unquote_it(buf, &n);
			strcpy(msg, buf);
			free(buf);
			ptr = (*dest == '#' || *dest == '&') ? fmt_sed_pub : fmt_sed_priv;
		}
		else
			ptr = fmt_sed_nokey;

		fmt_string = "%n%l%h%c%p%z%Z$";
		buf = format_msg(ptr, nick, login, host, dest, msg, ll, Win->server->read);
		if (buf) {
			fl_addto_browser(Win->chanwin->browser, buf);
			free(buf);
		}
		if (sflags & S_SED)
			addto_sender(sender, SS_SED);
		free(msg);  free(sender);
	}
}

static void do_action(void *data)
{
	if (!check_msg_hook(ACTION, Win - winstruct, "%s,%d %s",
			    Win->server->servname, Win->server->port, Msg))
	{
		char   *buf;
		char   *nick, *dest, *login, *host, *msg;
		char   *sender;
		size_t  ll;
		int     i;

		ll = strlen(Msg);

		sender = (char *) alloca(sizeof(char) * ll);

		if (sscanf(Msg, "%s", sender) < 1)
			return;
		msg = (char *) alloca(sizeof(char) * ll);
		dest = (char *) alloca(sizeof(char) * ll);

		ll = strlen(sender)+8;
		if (sscanf(Msg, "%*s PRIVMSG %s :\x1%*s %[^\x1\r\n]\x1\r\n", dest, msg) < 2) {
			if (gflags & VERBOSE_CTCP)
				say2(-1,1,"invalid CTCP action str: %s", Msg);
			return;
		}
		unspace(msg);
		if(*msg==0) return;
		nick = (char *) alloca(sizeof(char) * ll);
		login = (char *) alloca(sizeof(char) * ll);
		host = (char *) alloca(sizeof(char) * ll);

		i = sscanf(sender, "%[^!]!%[^@]@%s", nick, login, host);
		if (i < 2) {
			strcpy(login,"<unknown>");
			nick = sender;
		}
		else if (i < 3) {
			strcpy(host,"<unknown>");
		}
		buf = 0;
		if(*login=='~') strshift(login,1);
		fmt_string = "%n%l%h%c%p%z%Z$";
		buf = format_msg(
		 (*dest == '#' || *dest == '&') ? fmt_pub_action : fmt_priv_action,
		 nick, login, host, dest, msg, strlen(Msg), Win->server->read);
		if (buf)
			fl_addto_browser(Win->chanwin->browser, buf);
		free(buf);
		if (sflags & S_ACTION)
			addto_sender(sender, SS_ACTION);
	}
}

static void do_clientinfo(void *data)
{

	#if CTCP_more
	if (!check_msg_hook(CTCP_CLIENTINFO, Win - winstruct, "%s,%d %s",
			    Win->server->servname, Win->server->port, Msg))
	#endif
	 {
		char   *s, buf[512];
		char   *nick = (char *) alloca(sizeof(char) * strlen(Msg));
		char   *tt;

		if (sscanf(Msg, "%[^!]!", nick) < 1)
			return;
		s = (char *) malloc(sizeof(char) * (strlen(sula_GETTING) +
		       strlen  (sula_INFOSITE) + strlen(sula_EMAIL) + 200));

		if (!s)
			return;
		sprintf(s, "CLIENTINFO sula - a programmable X client \n"
			"Infoterial: " sula_INFOSITE " \n"
			"Email: " sula_EMAIL);
		tt = ctcp_quote_it(s, strlen(s));
		sprintf(buf, "NOTICE %s :%c%s%c\n", nick, 0x01, tt, 0x01);
		sendto_server(Win, buf);
		free(s);
		free(tt);
	}
}
static void do_source(void *data)
{
	#if CTCP_more
	if (!check_msg_hook(CTCP_SOURCE, Win - winstruct, "%s,%d %s",
			    Win->server->servname, Win->server->port, Msg))
	#endif
	 {
		char   *s, *tt, buf[512];
		char   *nick = (char *) alloca(sizeof(char) * strlen(Msg));

		if (sscanf(Msg, "%[^!]!", nick) < 1)
			return;
		s = (char *) malloc(sizeof(char) * (strlen(sula_GETTING) + 20));

		sprintf(s, "SOURCE " sula_GETTING);
		tt = ctcp_quote_it(s, strlen(s));
		sprintf(buf, "NOTICE %s :%c%s%c\n", nick, 0x01, tt, 0x01);
		sendto_server(Win, buf);
		free(s);
		free(tt);
	}
}

#define DCC_CHAT	0
#define DCC_SEND	1
#define DCC_ACCEPT	2
#define DCC_RESUME	3
#define DCC_RENAME	4
#define DCC_REJECT	5

static void parse_dcc_request(int which)
{

	char   *nick,  *email, *msg;
	int     ll;
	unsigned short port = 0;

	ll = strlen(Msg);
	msg = (char *) alloca(sizeof(char) * ll);
	sscanf(Msg, "%*s PRIVMSG %*s :\x1%*s %*s %[^\x1\r\n]\x1\r\n", msg);
	email = (char *) alloca(sizeof(char) * ll);
	nick = (char *) alloca(sizeof(char) * ll);
	if (sscanf(Msg, "%[^!]!%s", nick, email) < 2) {
		strcpy(nick,"#unknown#");
		strcpy(email,"#unknown#");
	}
	else
	if (*email == '~') strshift(email,1);
	switch (which) {
	    case DCC_SEND:
		    if (!check_msg_hook(DCC_SEND_REQUEST,
					Win - winstruct, "%s,%d %s",
			   Win->server->servname, Win->server->port, Msg)) {
			    char   *fname, *ext_block;
			    /*CheckSum */ u_long cksum = 0;
			    u_long  inaddr = 0;
			    u_long  fsize = 0;

			    ext_block = (char *) alloca(sizeof(char) * strlen(msg));

			    *ext_block = 0;
			    fname = (char *) alloca(sizeof(char) * strlen(msg));

			    *fname = 0;
			    sscanf(msg, "%s %lu %hu %lu %lu",
				   fname, &inaddr, &port, &fsize, &cksum);
			    if (*fname == 0 || port == 0) {
				    if (gflags & VERBOSE_CTCP)
					    fprintf(stderr,
						    "invalid DCC SEND Req. %s from %s\n", msg, nick);
				    break;
			    }
			    /* read extension block */
			    sscanf(Msg, "%*[^\x1]\x1%*[^\x1]\x1\x2%[^\x2]\x2", ext_block);
#if DEBUG
			    fprintf(stderr, "DCC SEND: extension block:%s (%d bytes)\n",
				    ext_block, strlen(ext_block));
#endif
			    /*if(*ext_block) sscanf(ext_block,"%lu", &cksum); */
			    if (sflags & S_FILESEND){
				    char *c=nextstr(msg, 1);
				    addto_sender(c, SS_FILESEND);
				    free(c);
			   }
			    receive_send_req(nick, email, fname, inaddr, port, fsize, cksum);
			    break;
		    }
	    case DCC_ACCEPT:
		    {
			    off_t   offset = 0;

			    /* offset is redundant: open in O_APPEND mode */
			    sscanf(msg, "%*s %hu %lu", &port, &offset);
			    if (port == 0) {
			      if (gflags & VERBOSE_CTCP)
				error(0,"incomplete req. from %s:%s\n", msg, nick);
			    }
			    else
				    accept_request(nick, email, port, offset);
			    break;
		    }
	    case DCC_RESUME:
		    {
			    off_t   offset = 0;
			    u_long  checksum = 0;

#if 1
			    char   *pp = malloc(sizeof(char) * ll), *ss = 0;

			    sscanf(msg, "%s %hu %lu", pp, &port, &offset);
			    checksum = strtoul(pp, &ss, 0);
			    if (*ss)
				    checksum = 0;
			    free(pp);
#else
			    sscanf(msg, "%lu %hu %lu", &checksum, &port, &offset);
#endif
			    if (port)
				    resume_request(nick, email, port, offset, checksum);
			    else if (gflags & VERBOSE_CTCP)
				    fprintf(stderr, "incomplete req. from %s:%s\n", nick, msg);
			    break;
		    }
	    case DCC_RENAME:
		    {
			    sscanf(msg, "%*s %hu", &port);
			    if (port == 0) {
				    if (gflags & VERBOSE_CTCP)
					    fprintf(stderr, "incomplete req. from %s:%s\n", msg, nick);
			    }
			    else
				    rename_request(nick, email, port);
			    break;
		    }
	    case DCC_REJECT:
		    {
			    sscanf(msg, "%*s %hu", &port);
			    reject_request(nick, email, port);
			    break;
		    }
	    case DCC_CHAT:
		    if (!check_msg_hook(DCC_CHAT_REQUEST,
					Win - winstruct, "%s,%d %s",
			   Win->server->servname, Win->server->port, Msg)) {
			    u_long  inaddr;

			    if (sscanf(msg, "%*s %lu %hu", &inaddr, &port) < 2) {
				    if (gflags & VERBOSE_CTCP)
					    fprintf(stderr, "bad DCC CHAT Req. %s\n", msg);
				    break;
			    }
			    if (sflags & S_FILEGET){
				    char *c=nextstr(msg, 1);
				    addto_sender(c, SS_FILEGET);free(c);
				}
			    receive_chat_req(nick, email, inaddr, port);
			    break;
		    }
	    default:
		    if (gflags & VERBOSE_CTCP)
			    fprintf(stderr, "unimplemented DCC type %d\n", which);
		    break;
	}			/* swicth(which) */
}

static int lookup_DCC_type(const char *name)
{
	typedef struct TDCC_request {
		short   nr;
		char   *name;
	} DCC_request;
	DCC_request request_type[] =
	{
	/* binary search, keep sorted by name!! */
		{DCC_ACCEPT, "ACCEPT"},
		{DCC_CHAT, "CHAT"},
		{DCC_REJECT, "REJECT"},
		{DCC_RENAME, "RENAME"},
		{DCC_RESUME, "RESUME"},
		{DCC_SEND, "SEND"},
		{-1, NULL}
	},     *p = request_type;
	register int low, high, mid, cond;

	low = 0;
	high = sizeof(request_type) / sizeof(DCC_request) - 1;
	while (p->name && low <= high) {
		mid = (low + high) / 2;
		if ((cond = strcasecmp(name, (p + mid)->name)) < 0)
			high = mid - 1;
		else if (cond > 0)
			low = mid + 1;
		else {
			parse_dcc_request((p + mid)->nr);
			return mid;
		}
	}
	return -1;
}

static void do_dcc(void *data)
{

	char   *cmd, *msg;
	int     ll;

	ll = strlen(Msg);
	msg = (char *) alloca(sizeof(char) * ll);
	cmd = (char *) alloca(sizeof(char) * ll);

	ll = sscanf(Msg, "%*s PRIVMSG %*s :\x1%*s %s %[^\x1\r\n]\x1\r\n",
		    cmd, msg);
	if (ll < 2) {
		if (gflags & VERBOSE_CTCP)
			error(0,"invalid DCC: %s", Msg);
	}
	else if (lookup_DCC_type(cmd) < 0)
		if (gflags & VERBOSE_CTCP)
			error(0,"unknown DCC cmd %s (%s)\n", cmd, Msg);
}

static void do_ping(void *data)
{
	#if CTCP_more
	if (!check_msg_hook(CTCP_PING, Win - winstruct, "%s,%d %s",
			   Win->server->servname, Win->server->port, Msg))
	#endif
	{
		char    s[256];
		char   *nick = (char *) alloca(sizeof(char) * strlen(Msg));
		char   *buf = (char *) alloca(sizeof(char) * strlen(Msg));

		if (sscanf(Msg, "%[^!]!%*s %*s %*s :\x1%*s %[^\x1]\x1", nick, buf) < 2)
			return;
		sprintf(s, "NOTICE %s :%cPING %s%c\n", nick, 0x01, buf, 0x01);

		sendto_server(Win, s);
		if (sflags & S_PING){
			char *c=nextstr(Msg, 1);
			addto_sender(c, SS_PING); free(c);
		}
	}

}

static void do_time(void *data)
{
	#if CTCP_more
	if (check_msg_hook(CTCP_TIME,
			   Win - winstruct, "%s,%d %s", Win->server->servname, Win->server->port, Msg))
		return;
	else 
	#endif
	{
		char    s[256];
		char   *nick;
		time_t  p;
		char   *t;

		nick = alloca(sizeof(char) * strlen(Msg));

		p = time(NULL);
		t = ctime(&p);
		if (sscanf(Msg, "%[^!]!", nick) < 1)
			return;
		*(t + strlen(t) - 1) = '\0';
		sprintf(s, "NOTICE %s :%cTIME %s%c\n", nick, 0x01, t, 0x01);
		if (sflags & S_CTCP_OTHER){
			char *c=nextstr(Msg, 1);
			addto_sender(c, SS_CTCP_OTHER); free(c);
		}
		sendto_server(Win, s);
	}
}

static void do_version(void *data)
{
       #if CTCP_more
	if (!check_msg_hook(CTCP_VERSION,
			    Win - winstruct, "%s,%d %s",
			    Win->server->servname, Win->server->port, Msg)) 
	#endif
	{
		char   *s = NULL, buf[512];
		char   *nick = NULL, *tt = NULL;

		struct utsname ut;

		nick = (char *) alloca(sizeof(char) * strlen(Msg));

		if (sscanf(Msg, "%[^!]!", nick) < 1)
			return;
		if (uname(&ut) < 0) {
			char   *s;
			static char *release = NULL;

			if (release == 0) {
				time_t  t = sula_RELEASE;

				release = ctime(&t);
				*(release + strlen(release) - 1) = 0;
			}
			s = malloc(sizeof(char) * (strlen(sula_DESCR) + strlen(sula_VERSION) +
						   strlen  (release) + strlen(Win->server->nick) + 100));

			sprintf(s, "VERSION %s, version %s (Release %s) \nDo"
				" \"CTCP %s HELP\" for more info.",
				sula_DESCR, sula_VERSION, release, Win->server->nick);
		}
		else {
			s = (char *) malloc(sizeof(char) * 
			          (strlen(sula_DESCR) + strlen(sula_VERSION) +
		              strlen  (Win->server->nick) + strlen(ut.sysname) + 
		                strlen(ut.release) + 100));

			if (!s)
				return;
			sprintf(s, "VERSION %s, v%s on %s %s. \nDo"
				" \"CTCP %s HELP\" for more info.",
				sula_DESCR, sula_VERSION, ut.sysname, ut.release, Win->server->nick);
		}
		tt = ctcp_quote_it(s, strlen(s));
		sprintf(buf, "NOTICE %s :%c%s%c\n", nick, 0x01, tt, 0x01);
		if (sflags & S_CTCP_OTHER){
			char *c=nextstr(Msg, 1);
			addto_sender(c, SS_CTCP_OTHER); free(c);
		}
		sendto_server(Win, buf);
		free(s);
		free(tt);
	}

}

static void do_ctcp_help(void *data)
{
	#if CTCP_more
	if (!check_msg_hook(CTCP_HELP,
			    Win - winstruct, "%s,%d %s",
			    Win->server->servname, Win->server->port, Msg))
	#endif
	{
#define LINEBUF	4096
		char    s[LINEBUF], *buf;
		static void listcmds(Cmd * p, char **buf);
		char   *nick = (char *) alloca(sizeof(char) * strlen(Msg));
		char   *cmd = (char *) alloca(sizeof(char) * strlen(Msg));

		s[0] = '\0';
		if (sscanf(Msg, "%[^!]!", nick) < 1)
			return;
		if (sscanf(Msg, "%*s %*s %*s %*s %[^\x1]\x1", cmd) == 1) {
			Cmd    *p;

			cmd = strtok(cmd, " \r\n");
			p = find_ctcp_cmd(cmd, 0);
			if (!p)
				sprintf(s, "NOTICE %s :%cHELP %s: No such command%c\n", nick, 0x01, cmd, 0x01);

			else if (p->help)
				sprintf(s, "NOTICE %s :%cHELP %s%c\n", nick, 0x01, p->help, 0x01);

			else
				sprintf(s, "NOTICE %s :%cHELP %s: No help for command%c\n", nick, 0x01, cmd, 0x01);

			sendto_server(Win, s);
			return;
		}
		buf = (char *) malloc(sizeof(char) * LINEBUF);
		if (!buf) {
			error(-1, "malloc error");
			return;
		}
		buf[0] = '\0';
		listcmds(ctcp_cmdTree, &buf);
		if (buf[0])
			sprintf(s, "NOTICE %s :%cHELP Help topics: %s%c\n", nick, 0x01, buf, 0x01);
		else
			sprintf(s, "NOTICE %s :%cHELP %s%c\n", nick, 0x01, "No help found for cmd", 0x01);
		if (sflags & S_CTCP_OTHER){
			char *c=nextstr(Msg, 1);
			addto_sender(c, SS_CTCP_OTHER); free(c);
		}
		sendto_server(Win, s);
		free(buf);
	}
}
static void listcmds(Cmd * p, char **s)
{
	if (p) {
		listcmds(p->left, s);
		 {
#warning 10/98 todo: use realloc(2)
			strcat(*s, p->name);
			strcat(*s, " ");
		}
		listcmds(p->right, s);
	}
}

static void do_userinfo(void *data)
{
	#if CTCP_more
	if (!check_msg_hook(CTCP_USERINFO,
			    Win - winstruct, "%s,%d %s",
			    Win->server->servname, Win->server->port, Msg))
	#endif
	{
		char   *s;
		char   *nick;

		nick = (char *) alloca(sizeof(char) * strlen(Msg));

		if (sscanf(Msg, "%[^!]!", nick) < 1)
			return;
		if (sflags & S_CTCP_OTHER)
		if (sflags & S_CTCP_OTHER){
			char *c=nextstr(Msg, 1);
			addto_sender(c, SS_CTCP_OTHER); free(c);
		}
		if (!(*userinfo))
			return;
		s = (char *) alloca(sizeof(char) * (strlen(userinfo) + strlen(nick) + 30));

		if (!s)
			return;
		sprintf(s, "NOTICE %s :%cUSERINFO %s%c\n", nick, 0x01, userinfo, 0x01);
		sendto_server(Win, s);
	}
}

static void do_finger(void *data)
{
	#if CTCP_more
	if (!check_msg_hook(CTCP_FINGER, Win - winstruct, "%s,%d %s",
			    Win->server->servname, Win->server->port, Msg))
	#endif
	{
		if (!(gflags & FINGER))
			return;
		else {
			char   *s;
			char   *nick;

			if (sflags & S_FINGER){
				s= nextstr(Msg, 1);
				addto_sender(s, SS_FINGER);
				free(s);
			}
			nick = (char *) alloca(sizeof(char) * strlen(Msg));
			if (sscanf(Msg, "%[^!]!", nick) < 1)
				return;
			s=NULL;
			if (*finger_reply) {
				s = alloca(sizeof(char) * (strlen(finger_reply) + 16));

				if (!s)
					return;
				sprintf(s, "FINGER %s", finger_reply);
			}
			else {
				struct passwd *pwd;
				FILE   *fp;

				pwd = getpwuid(getuid());
				if (!pwd)
					return;
#ifdef UTMP
				if (!(gflags & FINGER_UTMP))
					goto no_utmp;
				if ((fp = fopen(UTMP, "r")) != NULL) {
					struct utmp ut;
					char    name[UT_NAMESIZE + 1];
					int     found = 0;

					while (fread((char *) &ut, sizeof(ut), 1, fp) == 1) {
						if (ut.ut_user[0] == '\0')
							continue;
						strncpy(name, ut.ut_user, UT_NAMESIZE);
						if (!strcmp(pwd->pw_name, name)) {
							found = 1;
							break;
						}
					}
					fclose(fp);
					if (found) {
						char   *t;
						char    ut_host[UT_HOSTSIZE + 1];
						char    ut_line[UT_LINESIZE + 1];

						t = ctime(&ut.ut_time);
						*(t + strlen(t) - 1) = '\0';
						s = alloca(sizeof(char) * (strlen(pwd->pw_gecos) + strlen(t) +
									   strlen  (pwd->pw_name) + UT_HOSTSIZE + UT_NAMESIZE + 128 + strlen(host)
							   ));

						strncpy(ut_host, ut.ut_host, UT_HOSTSIZE);
						strncpy(ut_line, ut.ut_line, UT_LINESIZE);
						sprintf(s,
						 "FINGER Login:%s 	Name:%s\n"
					      "Host: %s From: %s Device:%s.\n"
					      "On since %s ",
						pwd->pw_name, pwd->pw_gecos,
						 host, ut_host, ut_line, t);
					}
				}
				else
					perror(UTMP);
			      no_utmp:
#endif 
				if (!s) {
					s = alloca(sizeof(char) * (strlen(pwd->pw_gecos) + strlen(pwd->pw_name) +
					strlen  (host) + 64));

					sprintf(s, "FINGER %s (%s) on host %s",
					 pwd->pw_name, pwd->pw_gecos, host);
				}
			}	/* if(*finger_reply) */
			if (s) {
				int     l = strlen(s);
				char   *tt = ctcp_quote_it(s, l);
				char   *b;
				b = alloca(sizeof(char) * (l + 16));

				sprintf(b, "NOTICE %s :%c%s%c\n", nick, 0x01, tt, 0x01);
				sendto_server(Win, b);
				free(tt);
			}
		}
	}
}

extern int load_ctcp_cmds(void)
{
	typedef struct TCtcpCmd {
		char   *name;
		char   *help;
		void    (*func) (void *);
	} CtcpCmd;
	CtcpCmd ctcp_cmds[] =
	{
		{"ACTION", "ACTION describes sender's actions or feelings", &do_action},
		{"CLIENTINFO", "CLIENTINFO  returns info about the client", &do_clientinfo},
	 {"DCC", "DCC request direct client to client connection", &do_dcc},
		{"FINGER", "FINGER gets system info about the recipient", &do_finger},
		{"HELP", "HELP gets info about a client's CTCP commands", &do_ctcp_help},
		{"PING", "PING sends something and expects it to be echoed", &do_ping},
		/*{"RELAY", "RELAY forwards a DCC request", &do_relay},*/
		{"SED", "SED marks simple(?) encryted data", &do_sed},
	 {"SOURCE", "SOURCE where to get a copy of the client", &do_source},
	 {"TIME", "TIME requests for the time on recipient host", &do_time},
		{"USERINFO", "USERINFO gets any info that user set for this purpose", &do_userinfo},
	    {"VERSION", "VERSION requests for client version", &do_version},

	};

	Cmd     cmd;
	int     nvars = sizeof(ctcp_cmds) / sizeof(CtcpCmd);
	int     i;

	for (i = 0; i < nvars; i++) {
		cmd.name = ctcp_cmds[i].name;
		cmd.help = ctcp_cmds[i].help;
		cmd.type = CMD_IN | CMD_CTCP;
		cmd.cmd.in = ctcp_cmds[i].func;
		insert_ctcp_cmd(&cmd);
	}
	return 0;

}

extern int ctcp_cmd(void)
{
	char   *str;
	 char   *buf, *sender;
	size_t  ll = strlen(Msg);
	
	buf = alloca(sizeof(char) * ll);
	if (sscanf(Msg, "%*s %*s %*s :\x1%[^\x1]\x1", buf) < 1)
		return 0;
	sender = (char *) alloca(sizeof(char) * ll);
	if (sscanf(Msg, "%s", sender) < 1)
		return 1;	/* sender */
	str = nextstr(buf, 1);	/* ctcp command */
	if (*str == 0)
	       {free(str); return 1;}
if ((!strcasecmp(str, "DCC") && ignored(sender, IG_DCC))
	    || (strcasecmp(str, "CTCP") && ignored(sender, IG_CTCP))
	   ) {
		if (gflags & INFORM_IGNORED) {
			char   *m, *nick, *email;

			ll = strlen(sender);
			nick = (char *) alloca(sizeof(char) * ll);
			email = (char *) alloca(sizeof(char) * ll);

			if (sscanf(sender, "%[^!]!%s", nick, email) < 2)
			       {free(str); return 1;}
			if(*email=='~') strshift(email,1);
			fmt_string = "%n%h%p$";
			m = format_msg(fmt_ignored, nick, email, str);
			if (!m)
				return 1;
			else {
				char   *s = (char *) alloca(sizeof(char) * (
				strlen(m) + strlen(nick) + 16));

				sprintf(s, "NOTICE %s :%s\n", nick, m);
				sendto_server(Win, s);
				free(m);
			}
		}
	       free(str);	return 1;
 }else
 {
	Cmd    *cmd;
	char *from = alloca(sizeof(char)*strlen(sender));
	if(sscanf(sender, "%*[^!]!%s",from)<1) return 1;
	if(*from=='~') strshift(from,1);
	if(sflags&FLOOD_CHECK && flooding(from,PRIVATE_FLOOD,FLOOD_CTCP) &&
	   check_msg_hook(PRIV_FLOOD,
	             Win-winstruct,"%s,%d %s",
	             Win->server->name,
	             Win->server->port,
	             Msg))
	    {free(str); return 1;}
	if (check_msg_hook(CTCP_ALL, Win - winstruct, "%s,%d %s",
			   Win->server->name, Win->server->port, Msg))
		return 1;
	cmd = find_ctcp_cmd(str, 0);
	if (!cmd) {
		if (!check_msg_hook(CTCP_UNKNOWN, Win - winstruct, "%s,%d %s",
				 Win->server->name, Win->server->port, Msg))
			if (gflags & VERBOSE_CTCP)
				error(0, "unknown ctcp %s received from %s: %s",
				      str, sender, buf);
		{free(str); return 1;}
	}
	free(str);
	if (gflags & VERBOSE_CTCP){
		str= nextstr(buf, 2);
		say2(-1, 1, PROMPT "CTCP %s %s received from %s",
		     cmd->name,str, sender);
		free(str);
	}
	if (cmd->type & CMD_IN) {
		(cmd->cmd.in) ((void *) 0);
	}
	else if (cmd->type & CMD_FD) {
		char   *buff;

		buff = (char *) malloc(sizeof(char) * (
		 ll  + strlen(Win->server->name) + 32));

		if (buff == NULL)
			error(-2, "malloc error");
		rmstr(buf,0); 
		/*
		   <id> __CTCP__ <yournick> <sender> <args>
		 */
		sprintf(buff, "%d __CTCP__ %s,%hu %s %s %s %s\n",
			Win - winstruct, Win->server->name,
			Win->server->port, Win->server->nick,
			sender, cmd->name, buf);
		write_to_sp(cmd->cmd.sp, buff);
		free(buff);
	}
#ifdef _GUILE
	else if (cmd->type & CMD_SCM) {
		char   *buff = alloca(sizeof(char) * (ll + 64));
		#define DAT_SIZE 128
		char data[DAT_SIZE+1];
		Scm_cmd_arg args;
		args.func=cmd->cmd.scm;
		sprintf(buff, "%s,%hu %s", Win->server->name,
			Win->server->port,
			Msg);
		args.args=gh_list(gh_str02scm(buff),
			 gh_long2scm(Win - winstruct),
			 SCM_UNDEFINED);
		
		if(snprintf(data,DAT_SIZE,"While processing CTCP cmd \"%s\"",(char *)buff)==-1)
		  {data[DAT_SIZE]=0;
	           strcpy(&data[DAT_SIZE-4],"...");
		   data[DAT_SIZE-1]='"';
                   }
		gh_catch(SCM_BOOL_T, &call_scm_command, &args,
		 (scm_catch_handler_t)&exception_handler, (void *)&data);
	}
#endif
	else
	    fprintf(stderr, "bug:unknown ctcp cmd type %d\n",
	     cmd->type & CMD_CTCP);
	return 1;
 }
}

/* -- originally from EPIC; 
   modified for sula by Tano Fotang 10/1998
   o return quoted string
   o don't quote "\n" ,"\t" etc

   * quote_it: This quotes the given string making it sendable via irc.  A
   * pointer to the length of the data is required and the data need not be
   * null terminated (it can contain nulls).  Returned is a malloced, null
   * terminated string.   
 */
char   *
        ctcp_quote_it(str, len)
     char   *str;
     int     len;
{
	char   *buffer;
	char   *ptr;
	register int i;

	buffer = (char *) malloc(sizeof(char) * (len + len));

	if (!buffer)
		return NULL;
	ptr = buffer;
	for (i = 0; i < len; i++) {
		switch (str[i]) {
		    case CTCP_DELIM_CHAR:
			    *(ptr++) = CTCP_QUOTE_CHAR;
			    *(ptr++) = 'a';
			    break;
		    case '\n':
			    *(ptr++) = CTCP_QUOTE_CHAR;
			    *(ptr++) = 'n';
			    break;
		    case '\r':
			    *(ptr++) = CTCP_QUOTE_CHAR;
			    *(ptr++) = 'r';
			    break;
		    case CTCP_QUOTE_CHAR:
		    /* sula requires this */
			    switch (str[i + 1]) {
				case 'n':
				case 't':
				case 'r':
				case '0':
				case CTCP_QUOTE_CHAR:
					*(ptr++) = str[i];
					break;
				default:
					*(ptr++) = CTCP_QUOTE_CHAR;
					*(ptr++) = CTCP_QUOTE_CHAR;
					break;
			    }
			    break;
		    case '\0':
			    *(ptr++) = CTCP_QUOTE_CHAR;
			    *(ptr++) = '0';
			    break;
		    default:
			    *(ptr++) = str[i];
			    break;
		}
	}
	*ptr = '\0';
	return buffer;
}
/* -- originally from EPIC; 
   modified for sula by Tano Fotang 10/98
   o s-str used as loop counter instead of an integer
   o pointer used instead of array str[i]
   o after \n, add multiline string to each new line
   o memcpy in place of the strange mem function in the original

   * NOTE: The CTCP_QUOTE_CHAR character is discarded if the succeeding char isnt
   * one of the following:
   * NUL, NL, CR, CTCP_QUOTE_CHAR --fotang.
 */
void    ctcp_unquote_it(char *str, int *len)
{
	char   *buffer;
	char   *ptr, c;
	char   *s;

	buffer = (char *) malloc(sizeof(char) * (*len + 16));

	if (!buffer)
		return;
	ptr = buffer;
	s = str;

	while (s - str < *len) {
		if ((c = *s++) == CTCP_QUOTE_CHAR) {
			switch (c = *s++) {
			    case CTCP_QUOTE_CHAR:
				    *(ptr++) = CTCP_QUOTE_CHAR;
				    break;
			    case 'a':
				    *(ptr++) = CTCP_DELIM_CHAR;
				    break;
			    case 'n':
				    {
					    char   *p = ml_nl;

					    *ptr++ = '\n';
					    if (p)
						    while (*p)
							    if (*p == '\\' && *(p + 1) == 't') {
								    *ptr++ = '\t';
								    p += 2;
							    }
							    else
								    *ptr++ = *p++;
				    }
				    break;
			    case 'r':
				    *(ptr++) = '\r';
				    break;
			    case 't':
				    *(ptr++) = '\t';
				    break;
			    case '0':
				    *(ptr++) = '\0';
				    break;
			    default:
				    *(ptr++) = c;
				    break;
			}
		}
		else
			*(ptr++) = c;
	}
	*ptr = 0;
	*len = ptr - buffer;
	memcpy(str, buffer, *len + 1);
}
/* -- from EPIC */
void    encrypt_str(str, len, key)
     char   *str, *key;
     int     len;
{
	int     key_len, key_pos, i;
	char    mix, tmp;

	key_len = strlen(key);
	key_pos = 0;
	/*    mix = key[key_len-1]; */
	mix = 0;
	for (i = 0; i < len; i++) {
		tmp = str[i];
		str[i] = mix ^ tmp ^ key[key_pos];
		mix ^= tmp;
		key_pos = (key_pos + 1) % key_len;
	}
	str[i] = (char) 0;
}

/* -- from ircII */
static void decrypt(str, len, key)
     char   *str, *key;
     int     len;
{
	int     key_len, key_pos, i;
	char    mix, tmp;

	key_len = strlen(key);
	key_pos = 0;
	/*    mix = key[key_len-1]; */
	mix = 0;
	for (i = 0; i < len; i++) {
		tmp = mix ^ str[i] ^ key[key_pos];
		str[i] = tmp;
		mix ^= tmp;
		key_pos = (key_pos + 1) % key_len;
	}
	str[i] = (char) 0;
}

