/* idle_cb.c - activities for an idle client ..........
 * distributed as part of _ s u l a _
 * 
 * Author: Tano Fotang, 1998
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License , or (at your option) 
 * any later version.
 * 
 * see the file COPYING for further information. 
 * 
 */



#include <forms.h>
#include <time.h>
#include "aux.h"
#include "sula.h"
#include "chanwin.h"
#include "global.h"
#include "dcc_gui.h"
#include "dcc.h"
#include "ignore.h"
#include "servers_gui.h"
#include "lag.h"
#include "idle_cb.h"
#include "cmd.h"
#include "flood.h"
#include "timer.h"

#ifndef NOLAG
static void check_server_lag(Server *p){
	if(p){
		Server *s=search_server_tree_byname(p->servname,p->port);
		assert(s);
		check_server_lag(p->left);
		if(s->lag) {
			struct timeval t;
			gettimeofday(&t,NULL);
			update_lag_display(s,
				t.tv_sec- s->lag->when.tv_sec+
		    (float)((t.tv_usec- s->lag->when.tv_usec)/ 1000000.0),
				  (struct timeval *)0);
		}
		check_server_lag(p->right);
	}
}
#endif
// clear flood list
//static jmp_buf	env;
#define DEL_CNT  50
static void update_flood(Flood *p,time_t t,
char *buf[DEL_CNT],int *n){
	if(p && *n<DEL_CNT ){
		update_flood(p->right,t,buf,n);
		if(t - p->at >flood_period){
		      buf[(*n)]=strdup(p->from);
		      if(buf[(*n)]==NULL) error(-2,"strdup error");
		      (*n)++	  ;
		}
		update_flood(p->right,t,buf,n);
	}
}
static void check_flood(pFlood *tree)
{
 
 //Flood **tree=priv?&floodTree:&publicfloodTree;
 if(tree){
     char *buf[DEL_CNT];
     int n=0;
     update_flood(*tree, time(NULL), buf,&n);
 while(n-->0){
   int res= clear_flood(buf[n],tree,0);
   assert(res==0);
   free(buf[n]);
 }
 }
}
// update ignore structures
static void check_ignore(void)
{
	Ignore *p = ignoreStart;
	time_t  t = time(NULL);

	while (p != ignoreEnd) {
		if (p->period == 0) {
			p = p->next;
			continue;
		}
		if ((p->starttime + 60 * p->period) < t) {
			Ignore *q = p->next;

			free(p->mask);
			if (q == ignoreEnd)
				ignoreEnd = p;
			else
				*p = *q;
			free(q);
		}
		else
			p = p->next;
	}
}
static void check_dcc_sent(int flag)
{

	time_t  t = time(NULL);

	if (flag) {
		ChatReqSent *p = chatsentStart;

		while (p != chatsentEnd)
			if (t > (p->starttime + dcc_timeout * 60))
				remove_sentchatreq(p->id);
			else
				p = p->next;
	}
	else {
		DCCFileS *q = filesendStart;

		while (q != filesendEnd)
			if (t > (q->starttime + dcc_timeout * 60))
				remove_file_offer(q->id);
			else
				q = q->next;
	}
}

static void check_dcc_recvd(int flag)
{
	time_t  t = time(NULL);

	if (flag) {
		ChatReqRecvd *q = chatreqStart;

		while (q != chatreqEnd)
			if (t > (q->starttime + dcc_timeout * 60)) {
				ChatReqRecvd *p = q->next;

				free(q->nick);
				free(q->email);
				if (p == chatreqEnd)
					chatreqEnd = q;
				else
					*q = *p;
				free(p);
				return;
			}
			else
				q = q->next;
	}
	else {
		DCCFileR *q = filegetStart;

		while (q != filegetEnd)
			if (t > (q->starttime + dcc_timeout * 60)) {
				DCCFileR *p = q->next;

				free(q->nick);
				free(q->email);
				free(q->filename);
				if (p == filegetEnd)
					filegetEnd = q;
				else
					*q = *p;
				free(p);
				return;
			}
			else
				q = q->next;
	}
}
extern ActiveDCC *aActiveDCC;
static void update_active_dcc_display(void)
{

	#if 1
	DCCFile *p = DCCfileStart;
	
	/*if(!(gflags&VERBOSE_CTCP)) return;*/
	while (p != DCCfileEnd){
		if(p->w>=0 && p->size>0 && p->read<p->size){
			char buf[20];
			float x=p->read * 100.0/p->size;
			struct timeval t;
			float dt;
			gettimeofday(&t,NULL);
			sprintf(buf,"%5.1f%%", x);
			fl_set_slider_value(aActiveDCC[p->w].form->slider,x);
			fl_set_object_label(aActiveDCC[p->w].form->slider,buf);
			dt=t.tv_sec-p->lasttime.tv_sec
			#if 1
			+
			(t.tv_usec-p->lasttime.tv_usec)/1000000.0
			#endif
			;
			if(dt==0) dt=1;
			sprintf(buf,"%10.1f kb/s",
			  (p->read- aActiveDCC[p->w].form->size->u_ldata)/(
			  dt * 1024));
			fl_set_object_label(aActiveDCC[p->w].form->speed,buf);
			aActiveDCC[p->w].form->size->u_ldata=p->read;
		}
		p = p->next;
	  }
	#endif
}
static void check_alarm(void)
{
register i;
time_t t=time(NULL);
for (i = 0; i < timer_count; i++)
  if(aTimer[i].type!=-1 && aTimer[i].flag&TIMER_ALARM && aTimer[i].expire.stop<=t){
       if(aTimer[i].type==ALARM_IN){
	    int ret=-999;
	    if(*(aTimer[i].cmd.cmd)=='/'){
	      if(*(aTimer[i].cmd.cmd+1) != 0)
	       ret=new_process_cmd(aTimer[i].cmd.cmd+1, -1);
	   }else
	       ret=new_process_cmd(aTimer[i].cmd.cmd, -1);
	       if(ret && ret!= -999){
	         fl_ringbell(0);
		 fprintf(stderr,"Timer:Unknown cmd? %s\n",aTimer[i].cmd.cmd);
		 }
	    free(aTimer[i].cmd.cmd);
	}
	#ifdef _GUILE
	else if (aTimer[i].type==ALARM_SCM){
	  #define DAT_SIZE 64
		char data[DAT_SIZE+1];
		Scm_cmd_arg args;
		args.func=aTimer[i].cmd.scm;
		args.args=gh_list(
		                gh_long2scm(aTimer[i].ref),
				gh_long2scm(aTimer[i].expire.stop),
				SCM_UNDEFINED);
		if(snprintf(data,DAT_SIZE,"Processing alarm %d at %lu",
		   aTimer[i].ref,aTimer[i].expire.stop)==-1)
		  data[DAT_SIZE]=0;
		scm_func[aTimer[i].cmd.scm].idx=-1;
		gh_catch(SCM_BOOL_T, &call_scm_command, &args,
		 (scm_catch_handler_t)&exception_handler, (void *)&data);
		 
	}
	#endif
	aTimer[i].type=-1;
        aTimer[i].expire.stop=0;
  }

}

static int clock_expired(struct tm *then, struct tm *now)
{
#if 1
//method 1
 return(
 (now->tm_year > then->tm_year ||
 (now->tm_year==then->tm_year && (
    now->tm_mon>then->tm_mon || 
      (now->tm_mon==then->tm_mon && (
         now->tm_mday>then->tm_mday ||
	   (now->tm_mday==then->tm_mday && (
	     now->tm_hour>then->tm_hour ||
	       (now->tm_hour==then->tm_hour && (
	         now->tm_min>then->tm_min ||
		   (now->tm_min==then->tm_min && 
		     now->tm_sec>=then->tm_sec)))))))))));
#else
//method 2
      time_t t1,t2;
      t1=mktime(now);
      t2=mktime(then);
      if(t1==-1 || t2==-1) return 1; // error-- ack, assume clock is overdue
      return(t1>=t2);
#endif
/*
(mtf 15.11.98)
Time[secs] required for 20000 runs:
+---------------+------------+------------+------------+
| 		|   user     |   system   |   real     |
+---------------+------------+------------+------------+
|       method1 |       0.01 |       0.00 |       0.01 |
+---------------+------------+------------+------------+
|       method2 |      31.71 |       0.00 |      32.67 |
+---------------+------------+------------+------------+
*/		 
}
static void check_clock(void)
{
register i;
struct tm *now;
time_t t=time(NULL);
now=localtime(&t);
for (i = 0; i < timer_count; i++)
  if(aTimer[i].type!=-1 && aTimer[i].flag&TIMER_CLOCK)
    if(clock_expired(aTimer[i].expire.date, now)){
       if(aTimer[i].type==ALARM_IN){
	    int ret=-99;
	    if(*(aTimer[i].cmd.cmd)=='/'){
	      if(*(aTimer[i].cmd.cmd+1) != 0)
	       ret=new_process_cmd(aTimer[i].cmd.cmd+1, -1);
	   }else
	       ret=new_process_cmd(aTimer[i].cmd.cmd, -1);
	       if(ret && ret!= -99){
	         fl_ringbell(0);
		 fprintf(stderr,"Clock:Unknown cmd? %s\n",aTimer[i].cmd.cmd);
		 }
	    free(aTimer[i].cmd.cmd);
	}
	#ifdef _GUILE
	else if (aTimer[i].type==ALARM_SCM){
	  #define DAT_SIZE 64
		char data[DAT_SIZE+1];
		Scm_cmd_arg args;
		args.func=aTimer[i].cmd.scm;
		args.args=gh_list(
		                gh_long2scm(aTimer[i].ref),
				gh_list(
				   gh_long2scm(aTimer[i].expire.date->tm_sec),
				   gh_long2scm(aTimer[i].expire.date->tm_min),
				   gh_long2scm(aTimer[i].expire.date->tm_hour),
				   gh_long2scm(aTimer[i].expire.date->tm_mday),
				   gh_long2scm(aTimer[i].expire.date->tm_mon),
				   gh_long2scm(aTimer[i].expire.date->tm_year),
				   SCM_UNDEFINED),
				SCM_UNDEFINED);
		if(snprintf(data,DAT_SIZE,"Processing clock %d at %lu",
		   aTimer[i].ref,aTimer[i].expire.stop)==-1)
		  data[DAT_SIZE]=0;
		scm_func[aTimer[i].cmd.scm].idx=-1;
		gh_catch(SCM_BOOL_T, &call_scm_command, &args,
		 (scm_catch_handler_t)&exception_handler, (void *)&data);
	}
	#endif
	aTimer[i].type=-1;
	free(aTimer[i].expire.date);
  }
}

extern int idle_cb(XEvent * ev, void *junk)
{
	static int flag = 0;

	flag %= 13;
	switch (++flag) {
	    case 1:
		    check_ignore();
		    break;
	    case 2:
		    check_dcc_sent(0);
		    break;
	    case 3:
		    check_dcc_sent(1);
		    break;
	    case 4:
		    check_dcc_recvd(0);
		    break;
	    case 5:
		    check_dcc_recvd(1);
		    break;
	    case 6:
		    if(servFdTree)
		    if (ignore_form && (time(NULL) % 60) == 0) {
			    static char i = 0;

			    i %= 3;	/* kludge to stop flashing on display */
			    if ((++i) == 1)
				    fl_call_object_callback(ignore_form->update);
		    }
		    break;
	    #ifndef NOLAG
	    case 7:
		send_lag_msg();
		break;
	    case 8:
		check_server_lag(servFdTree); /* update display */
		break;
	   #endif
	    case 9:
		    update_active_dcc_display();
		    break;
	    case 10:
	    	if(servFdTree)
		     check_flood(&floodTree);
		break;
	    case 11:
		 if(servFdTree)
		     check_flood(&publicfloodTree);
		break;
	    case 12:
	         if(aTimer)
		    check_alarm();
		break;
	    case 13:
	         if(aTimer)
		    check_clock();
		break;
	    default:
		    break;
	}
	return 0;
}
