/*
 * timer.c - create alarms and clocks
 * distributed as part of _ s u l a _
 * 
 * (C) 1998 Tano Fotang
 * 
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License , or (at your option) 
 * any later version.
 * see the file COPYING for further information.

 */


#include "aux.h"
#include "timer.h"
#include "util.h"
#include "scheme.h"
extern void error(int code, const char *fmt,...);

#define TIMER_ALLOC 5

static int timer_alloc(void)
{
	int     i;

	if (aTimer == NULL)
		aTimer = malloc(TIMER_ALLOC * sizeof(Timer));
	else
		aTimer = realloc(aTimer, (timer_count + TIMER_ALLOC) * sizeof(Timer));
	if (aTimer == NULL)
		return errno;
	for (i = timer_count; i < timer_count + TIMER_ALLOC; i++)
		aTimer[i].type = -1;
	timer_count += TIMER_ALLOC;
	return 0;
}
int add_timer(Timer *t, void *data)
{
	register i;

	if (aTimer == NULL && timer_alloc())
		return -1;
	do {
		for (i = 0; i < timer_count; i++)
			if (aTimer[i].type == -1) {
				if(t->type==ALARM_IN){
				   aTimer[i].cmd.cmd=strdup(t->cmd.cmd);
				   if(aTimer[i].cmd.cmd==NULL)
				       return -1;
				}
				else
				#ifdef _GUILE
				if(t->type==ALARM_SCM)
				   aTimer[i].cmd.scm=t->cmd.scm;
				else
				#endif
				error(2,"[%s %d] Unknown alarm type %d. Bye.",
				__FILE__,__FILE__, t->type);
				aTimer[i].type= t->type;
				aTimer[i].flag=t->flag;
				aTimer[i].ref = t->ref;
				if(t->flag&TIMER_ALARM)
				   aTimer[i].expire.stop = time(NULL) + *((time_t *)data);
				else
				if(t->flag&TIMER_CLOCK){
				   aTimer[i].expire.date=(struct tm *)malloc(sizeof(struct tm));
				   memcpy(aTimer[i].expire.date,data,sizeof(struct tm));
				}
				else
				   error(2,"[%s %d] Unknown timer type %d. Bye.",
				__FILE__,__FILE__, t->flag);
				return i;
			}
		if (timer_alloc())
			return -1;
	} while (1);
}

void del_timer(int ref, char type, char flag)
{
 register i;
 if(type==-1) return;
 for(i=0;i<timer_count;i++)
   if(aTimer[i].ref==ref && aTimer[i].flag&flag && aTimer[i].type==type){
       if(aTimer[i].type == ALARM_IN){
         free(aTimer[i].cmd.cmd);
	 aTimer[i].expire.stop=0;
       }
       #ifdef _GUILE
       else if(aTimer[i].type == ALARM_SCM)
            scm_func[aTimer[i].cmd.scm].idx=-1;
	#endif
	if(flag&TIMER_CLOCK)
	    free(aTimer[i].expire.date);
       aTimer[i].type = -1;
  }
}

